# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle  # nosec  # pylint: disable=import_pickle

import cloudpickle

from .... import opcodes
from ....core import OutputType, recursive_tile
from ....core.context import get_context
from ....core.operand import OperandStage, MergeDictOperand
from ....serialization.serializables import (
    KeyField,
    BytesField,
    DictField,
    Int32Field,
    Float32Field,
    FunctionField,
)
from ....utils import has_unknown_shape


class StatsModelsTrain(MergeDictOperand):
    _op_type_ = opcodes.STATSMODELS_TRAIN

    _exog = KeyField("exog")  # exogenous
    _endog = KeyField("endog")  # endogenous

    _num_partitions = Int32Field("num_partitions")
    _partition_id = Int32Field("partition_id")
    _factor = Float32Field("factor")
    _model_class = BytesField(
        "model_class", on_serialize=cloudpickle.dumps, on_deserialize=cloudpickle.loads
    )
    _init_kwds = DictField("init_kwds")
    _fit_kwds = DictField("fit_kwds")
    _estimation_method = FunctionField("estimation_method")
    _estimation_kwds = DictField("estimation_kwds")
    _join_method = FunctionField("join_method")
    _join_kwds = DictField("join_kwds")
    _results_class = BytesField(
        "results_class",
        on_serialize=cloudpickle.dumps,
        on_deserialize=cloudpickle.loads,
    )
    _results_kwds = DictField("results_kwds")

    def __init__(
        self,
        exog=None,
        endog=None,
        num_partitions=None,
        partition_id=None,
        factor=None,
        model_class=None,
        init_kwds=None,
        fit_kwds=None,
        estimation_method=None,
        estimation_kwds=None,
        join_method=None,
        join_kwds=None,
        results_class=None,
        results_kwds=None,
        **kw
    ):
        super().__init__(
            _exog=exog,
            _endog=endog,
            _num_partitions=num_partitions,
            _partition_id=partition_id,
            _factor=factor,
            _model_class=model_class,
            _init_kwds=init_kwds,
            _fit_kwds=fit_kwds,
            _estimation_method=estimation_method,
            _estimation_kwds=estimation_kwds,
            _join_method=join_method,
            _join_kwds=join_kwds,
            _results_class=results_class,
            _results_kwds=results_kwds,
            **kw
        )

    @property
    def exog(self):
        return self._exog

    @property
    def endog(self):
        return self._endog

    @property
    def num_partitions(self):
        return self._num_partitions

    @property
    def partition_id(self):
        return self._partition_id

    @property
    def factor(self):
        return self._factor

    @property
    def model_class(self):
        return self._model_class

    @property
    def init_kwds(self) -> dict:
        return self._init_kwds

    @property
    def fit_kwds(self) -> dict:
        return self._fit_kwds

    @property
    def estimation_method(self):
        return self._estimation_method

    @property
    def estimation_kwds(self) -> dict:
        return self._estimation_kwds

    @property
    def join_method(self):
        return self._join_method

    @property
    def join_kwds(self) -> dict:
        return self._join_kwds

    @property
    def results_class(self):
        return self._results_class

    @property
    def results_kwds(self) -> dict:
        return self._results_kwds

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(inputs)
        self._exog = next(inputs_iter)
        self._endog = next(inputs_iter)

    def __call__(self, exog, endog):
        self._output_types = [OutputType.object]
        return self.new_tileable([exog, endog])

    @classmethod
    def tile(cls, op: "StatsModelsTrain"):
        if op.factor is not None:
            ctx = get_context()
            cluster_cpu_count = ctx.get_total_n_cpu()
            assert cluster_cpu_count > 0
            num_partitions = int(cluster_cpu_count * op.factor)
        else:
            num_partitions = op.num_partitions

        if has_unknown_shape(op.exog, op.endog):
            yield

        exog = op.exog
        if exog.ndim > 1 and exog.chunk_shape[1] > 1:
            exog = exog.rechunk({1: exog.shape[1]})
        exog = yield from recursive_tile(exog.rebalance(num_partitions=num_partitions))
        endog = yield from recursive_tile(
            op.endog.rebalance(num_partitions=num_partitions)
        )

        assert len(exog.chunks) == len(endog.chunks)

        # generate map stage
        map_chunks = []
        for part_id, (exog_chunk, endog_chunk) in enumerate(
            zip(exog.chunks, endog.chunks)
        ):
            new_op = op.copy().reset_key()
            new_op._factor = None
            new_op._partition_id = part_id
            new_op._num_partitions = num_partitions
            new_op.stage = OperandStage.map

            map_chunks.append(
                new_op.new_chunk([exog_chunk, endog_chunk], index=exog_chunk.index)
            )

        # generate combine (join) stage
        new_op = op.copy().reset_key()
        new_op._factor = None
        new_op._num_partitions = num_partitions
        new_op.stage = OperandStage.combine

        combine_chunk = new_op.new_chunk(map_chunks, index=(0,))

        # generate tileable
        new_op = op.copy().reset_key()
        return new_op.new_tileables(op.inputs, chunks=[combine_chunk])

    @classmethod
    def _execute_map(cls, ctx, op: "StatsModelsTrain"):
        endog = ctx[op.endog.key]
        exog = ctx[op.exog.key]

        # code from statsmodels.base.distributed_estimation::_helper_fit_partition
        model = op.model_class(endog, exog, **op.init_kwds)
        results = op.estimation_method(
            model,
            op.partition_id,
            op.num_partitions,
            fit_kwds=op.fit_kwds,
            **op.estimation_kwds
        )
        ctx[op.outputs[0].key] = pickle.dumps(results)

    @classmethod
    def _execute_combine(cls, ctx, op: "StatsModelsTrain"):
        # code from statsmodels.base.distributed_estimation::DistributedModel.fit
        results_list = [pickle.loads(ctx[inp.key]) for inp in op.inputs]  # nosec
        params = op.join_method(results_list, **op.join_kwds)
        res_mod = op.model_class([0], [0], **op.init_kwds)
        result = op.results_class(res_mod, params, **op.results_kwds)

        ctx[op.outputs[0].key] = pickle.dumps(result)

    @classmethod
    def execute(cls, ctx, op: "StatsModelsTrain"):
        if op.merge:  # pragma: no cover
            super().execute(ctx, op)
        elif op.stage == OperandStage.combine:
            cls._execute_combine(ctx, op)
        else:
            cls._execute_map(ctx, op)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import AnyField
from ..utils import gen_random_seeds
from .core import TensorRandomOperandMixin, handle_array, TensorDistribution


class TensorZipf(TensorDistribution, TensorRandomOperandMixin):
    _input_fields_ = ["a"]
    _op_type_ = OperandDef.RAND_ZIPF

    _fields_ = "a", "size"
    a = AnyField("a")
    _func_name = "zipf"

    def __call__(self, a, chunk_size=None):
        return self.new_tensor([a], None, raw_chunk_size=chunk_size)


def zipf(random_state, a, size=None, chunk_size=None, gpu=None, dtype=None):
    r"""
    Draw samples from a Zipf distribution.

    Samples are drawn from a Zipf distribution with specified parameter
    `a` > 1.

    The Zipf distribution (also known as the zeta distribution) is a
    continuous probability distribution that satisfies Zipf's law: the
    frequency of an item is inversely proportional to its rank in a
    frequency table.

    Parameters
    ----------
    a : float or array_like of floats
        Distribution parameter. Should be greater than 1.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``a`` is a scalar. Otherwise,
        ``mt.array(a).size`` samples are drawn.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    dtype : data-type, optional
      Data-type of the returned tensor.

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized Zipf distribution.

    See Also
    --------
    scipy.stats.zipf : probability density function, distribution, or
        cumulative density function, etc.

    Notes
    -----
    The probability density for the Zipf distribution is

    .. math:: p(x) = \frac{x^{-a}}{\zeta(a)},

    where :math:`\zeta` is the Riemann Zeta function.

    It is named for the American linguist George Kingsley Zipf, who noted
    that the frequency of any word in a sample of a language is inversely
    proportional to its rank in the frequency table.

    References
    ----------
    .. [1] Zipf, G. K., "Selected Studies of the Principle of Relative
           Frequency in Language," Cambridge, MA: Harvard Univ. Press,
           1932.

    Examples
    --------
    Draw samples from the distribution:

    >>> import mars.tensor as mt

    >>> a = 2. # parameter
    >>> s = mt.random.zipf(a, 1000)

    Display the histogram of the samples, along with
    the probability density function:

    >>> import matplotlib.pyplot as plt
    >>> from scipy import special

    Truncate s values at 50 so plot is interesting:

    >>> count, bins, ignored = plt.hist(s[s<50].execute(), 50, normed=True)
    >>> x = mt.arange(1., 50.)
    >>> y = x**(-a) / special.zetac(a)
    >>> plt.plot(x.execute(), (y/mt.max(y)).execute(), linewidth=2, color='r')
    >>> plt.show()
    """
    if dtype is None:
        dtype = np.random.RandomState().zipf(handle_array(a), size=(0,)).dtype

    size = random_state._handle_size(size)
    seed = gen_random_seeds(1, random_state.to_numpy())[0]
    op = TensorZipf(size=size, seed=seed, gpu=gpu, dtype=dtype)
    return op(a, chunk_size=chunk_size)



# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..oscar.supervisor import SupervisorCommandRunner
from .core import K8SServiceMixin


class K8SSupervisorCommandRunner(K8SServiceMixin, SupervisorCommandRunner):
    async def start_services(self):
        await super().start_services()
        await self.start_readiness_server()

    async def stop_services(self):
        await self.stop_readiness_server()
        await super().stop_services()


main = K8SSupervisorCommandRunner()

if __name__ == "__main__":  # pragma: no branch
    main()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .map import series_map, index_map
from .to_gpu import to_gpu
from .to_cpu import to_cpu
from .rechunk import rechunk
from .describe import describe
from .apply import df_apply, series_apply
from .transform import df_transform, series_transform
from .isin import series_isin, df_isin
from .cut import cut
from .qcut import qcut
from .shift import shift, tshift
from .diff import df_diff, series_diff
from .value_counts import value_counts
from .astype import astype, index_astype
from .drop import df_drop, df_pop, series_drop, index_drop
from .drop_duplicates import (
    df_drop_duplicates,
    series_drop_duplicates,
    index_drop_duplicates,
)
from .duplicated import df_duplicated, series_duplicated, index_duplicated
from .melt import melt
from .memory_usage import df_memory_usage, series_memory_usage, index_memory_usage
from .select_dtypes import select_dtypes
from .map_chunk import map_chunk
from .cartesian_chunk import cartesian_chunk
from .rebalance import rebalance
from .stack import stack
from .explode import df_explode, series_explode
from .eval import df_eval, df_query
from .check_monotonic import (
    check_monotonic,
    is_monotonic,
    is_monotonic_increasing,
    is_monotonic_decreasing,
)
from .pct_change import pct_change
from .transpose import transpose


def _install():
    from ..core import DATAFRAME_TYPE, SERIES_TYPE, INDEX_TYPE
    from .standardize_range_index import ChunkStandardizeRangeIndex
    from .string_ import _string_method_to_handlers
    from .datetimes import _datetime_method_to_handlers
    from .accessor import StringAccessor, DatetimeAccessor, CachedAccessor

    for t in DATAFRAME_TYPE:
        setattr(t, "to_gpu", to_gpu)
        setattr(t, "to_cpu", to_cpu)
        setattr(t, "rechunk", rechunk)
        setattr(t, "describe", describe)
        setattr(t, "apply", df_apply)
        setattr(t, "transform", df_transform)
        setattr(t, "isin", df_isin)
        setattr(t, "shift", shift)
        setattr(t, "tshift", tshift)
        setattr(t, "diff", df_diff)
        setattr(t, "astype", astype)
        setattr(t, "drop", df_drop)
        setattr(t, "pop", df_pop)
        setattr(
            t, "__delitem__", lambda df, items: df_drop(df, items, axis=1, inplace=True)
        )
        setattr(t, "drop_duplicates", df_drop_duplicates)
        setattr(t, "duplicated", df_duplicated)
        setattr(t, "melt", melt)
        setattr(t, "memory_usage", df_memory_usage)
        setattr(t, "select_dtypes", select_dtypes)
        setattr(t, "map_chunk", map_chunk)
        setattr(t, "cartesian_chunk", cartesian_chunk)
        setattr(t, "rebalance", rebalance)
        setattr(t, "stack", stack)
        setattr(t, "explode", df_explode)
        setattr(t, "eval", df_eval)
        setattr(t, "query", df_query)
        setattr(t, "pct_change", pct_change)
        setattr(t, "transpose", transpose)

    for t in SERIES_TYPE:
        setattr(t, "to_gpu", to_gpu)
        setattr(t, "to_cpu", to_cpu)
        setattr(t, "rechunk", rechunk)
        setattr(t, "map", series_map)
        setattr(t, "describe", describe)
        setattr(t, "apply", series_apply)
        setattr(t, "transform", series_transform)
        setattr(t, "isin", series_isin)
        setattr(t, "shift", shift)
        setattr(t, "tshift", tshift)
        setattr(t, "diff", series_diff)
        setattr(t, "value_counts", value_counts)
        setattr(t, "astype", astype)
        setattr(t, "drop", series_drop)
        setattr(t, "drop_duplicates", series_drop_duplicates)
        setattr(t, "duplicated", series_duplicated)
        setattr(t, "memory_usage", series_memory_usage)
        setattr(t, "map_chunk", map_chunk)
        setattr(t, "cartesian_chunk", cartesian_chunk)
        setattr(t, "rebalance", rebalance)
        setattr(t, "explode", series_explode)
        setattr(t, "check_monotonic", check_monotonic)
        setattr(t, "is_monotonic", property(fget=is_monotonic))
        setattr(t, "is_monotonic_increasing", property(fget=is_monotonic_increasing))
        setattr(t, "is_monotonic_decreasing", property(fget=is_monotonic_decreasing))
        setattr(t, "pct_change", pct_change)

    for t in INDEX_TYPE:
        setattr(t, "map", index_map)
        setattr(t, "rechunk", rechunk)
        setattr(t, "rebalance", rebalance)
        setattr(t, "drop", index_drop)
        setattr(t, "drop_duplicates", index_drop_duplicates)
        setattr(t, "duplicated", index_duplicated)
        setattr(t, "memory_usage", index_memory_usage)
        setattr(t, "astype", index_astype)
        setattr(t, "value_counts", value_counts)
        setattr(t, "check_monotonic", check_monotonic)
        setattr(t, "is_monotonic", property(fget=is_monotonic))
        setattr(t, "is_monotonic_increasing", property(fget=is_monotonic_increasing))
        setattr(t, "is_monotonic_decreasing", property(fget=is_monotonic_decreasing))

    for method in _string_method_to_handlers:
        if not hasattr(StringAccessor, method):
            StringAccessor._register(method)

    for method in _datetime_method_to_handlers:
        if not hasattr(DatetimeAccessor, method):
            DatetimeAccessor._register(method)

    for series in SERIES_TYPE:
        series.str = CachedAccessor("str", StringAccessor)
        series.dt = CachedAccessor("dt", DatetimeAccessor)


_install()
del _install



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from .core import TensorFFTMixin, validate_fft, TensorHermitianFFT


class TensorIHFFT(TensorHermitianFFT, TensorFFTMixin):
    _op_type_ = OperandDef.IHFFT

    def __init__(self, n=None, axis=-1, norm=None, **kw):
        super().__init__(_n=n, _axis=axis, _norm=norm, **kw)

    @classmethod
    def _get_shape(cls, op, shape):
        new_shape = list(shape)
        shape = op.n if op.n is not None else shape[op.axis]
        if shape % 2 == 0:
            shape = (shape // 2) + 1
        else:
            shape = (shape + 1) // 2
        new_shape[op.axis] = shape
        return tuple(new_shape)


def ihfft(a, n=None, axis=-1, norm=None):
    """
    Compute the inverse FFT of a signal that has Hermitian symmetry.

    Parameters
    ----------
    a : array_like
        Input tensor.
    n : int, optional
        Length of the inverse FFT, the number of points along
        transformation axis in the input to use.  If `n` is smaller than
        the length of the input, the input is cropped.  If it is larger,
        the input is padded with zeros. If `n` is not given, the length of
        the input along the axis specified by `axis` is used.
    axis : int, optional
        Axis over which to compute the inverse FFT. If not given, the last
        axis is used.
    norm : {None, "ortho"}, optional
        Normalization mode (see `numpy.fft`). Default is None.

    Returns
    -------
    out : complex Tensor
        The truncated or zero-padded input, transformed along the axis
        indicated by `axis`, or the last one if `axis` is not specified.
        The length of the transformed axis is ``n//2 + 1``.

    See also
    --------
    hfft, irfft

    Notes
    -----
    `hfft`/`ihfft` are a pair analogous to `rfft`/`irfft`, but for the
    opposite case: here the signal has Hermitian symmetry in the time
    domain and is real in the frequency domain. So here it's `hfft` for
    which you must supply the length of the result if it is to be odd:

    * even: ``ihfft(hfft(a, 2*len(a) - 2) == a``, within roundoff error,
    * odd: ``ihfft(hfft(a, 2*len(a) - 1) == a``, within roundoff error.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> spectrum = mt.array([ 15, -4, 0, -1, 0, -4])
    >>> mt.fft.ifft(spectrum).execute()
    array([ 1.+0.j,  2.-0.j,  3.+0.j,  4.+0.j,  3.+0.j,  2.-0.j])
    >>> mt.fft.ihfft(spectrum).execute()
    array([ 1.-0.j,  2.-0.j,  3.-0.j,  4.-0.j])

    """
    a = astensor(a)
    validate_fft(a, axis=axis, norm=norm)
    op = TensorIHFFT(n=n, axis=axis, norm=norm, dtype=np.dtype(np.complex_))
    return op(a)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess
import weakref
from urllib.parse import urlparse
from typing import List, Dict, Union, Tuple, Iterator, BinaryIO, TextIO

import pyarrow as pa
from pyarrow.fs import (
    FileSystem as ArrowFileSystem,
    LocalFileSystem as ArrowLocalFileSystem,
    HadoopFileSystem as ArrowHadoopFileSystem,
    FileSelector,
    FileInfo,
    FileType,
)

from ...utils import implements, stringify_path
from .core import FileSystem, path_type


__all__ = ("ArrowBasedLocalFileSystem", "HadoopFileSystem")


# When pyarrow.fs.FileSystem gc collected,
# the underlying connection will be closed,
# so we hold the reference to make sure
# FileSystem will not be gc collected before file object
_file_to_filesystems = weakref.WeakKeyDictionary()


class ArrowBasedFileSystem(FileSystem):
    """
    FileSystem implemented with arrow fs API (>=2.0.0).
    """

    def __init__(self, arrow_fs: ArrowFileSystem, sequential_read=False):
        self._arrow_fs = arrow_fs
        # for open('rb'), open a sequential reading only or not
        self._sequential_read = sequential_read

    @staticmethod
    def _process_path(path):
        return stringify_path(path)

    @implements(FileSystem.cat)
    def cat(self, path: path_type) -> bytes:
        path = self._process_path(path)
        file: pa.NativeFile = self._arrow_fs.open_input_stream(path)
        return file.read()

    @implements(FileSystem.ls)
    def ls(self, path: path_type) -> List[path_type]:
        path = self._process_path(path)
        file_selector: FileSelector = FileSelector(path)
        paths = []
        for file_info in self._arrow_fs.get_file_info(file_selector):
            paths.append(file_info.path)
        return paths

    def _get_file_info(self, path: path_type) -> FileInfo:
        path = self._process_path(path)
        file_info: FileInfo = self._arrow_fs.get_file_info([path])[0]
        return file_info

    @implements(FileSystem.delete)
    def delete(self, path: path_type, recursive: bool = False):
        path = self._process_path(path)
        info = self._get_file_info(path)
        if info.is_file:
            self._arrow_fs.delete_file(path)
        elif info.type == FileType.Directory:
            if not recursive and len(self.ls(path)) > 0:
                raise OSError(f"[Errno 66] Directory not empty: '{path}'")
            self._arrow_fs.delete_dir(path)
        else:  # pragma: no cover
            raise TypeError(f"path({path}) to delete must be a file or directory")

    @implements(FileSystem.rename)
    def rename(self, path: path_type, new_path: path_type):
        path = self._process_path(path)
        new_path = self._process_path(new_path)
        self._arrow_fs.move(path, new_path)

    @implements(FileSystem.stat)
    def stat(self, path: path_type) -> Dict:
        path = self._process_path(path)
        info = self._get_file_info(path)
        stat = dict(name=path, size=info.size, modified_time=info.mtime_ns / 1e9)
        if info.type == FileType.File:
            stat["type"] = "file"
        elif info.type == FileType.Directory:
            stat["type"] = "directory"
        else:  # pragma: no cover
            stat["type"] = "other"
        return stat

    @implements(FileSystem.mkdir)
    def mkdir(self, path: path_type, create_parents: bool = True):
        path = self._process_path(path)
        self._arrow_fs.create_dir(path, recursive=create_parents)

    @implements(FileSystem.isdir)
    def isdir(self, path: path_type) -> bool:
        path = self._process_path(path)
        info = self._get_file_info(path)
        return info.type == FileType.Directory

    @implements(FileSystem.isfile)
    def isfile(self, path: path_type) -> bool:
        path = self._process_path(path)
        info = self._get_file_info(path)
        return info.is_file

    @implements(FileSystem._isfilestore)
    def _isfilestore(self) -> bool:
        return True

    @implements(FileSystem.exists)
    def exists(self, path: path_type):
        path = self._process_path(path)
        info = self._get_file_info(path)
        return info.type != FileType.NotFound

    @implements(FileSystem.open)
    def open(self, path: path_type, mode: str = "rb") -> Union[BinaryIO, TextIO]:
        path = self._process_path(path)
        is_binary = mode.endswith("b")
        if not is_binary:  # pragma: no cover
            raise ValueError(
                f"mode can only be binary for arrow based filesystem, got {mode}"
            )
        mode = mode.rstrip("b")
        if mode == "w":
            file = self._arrow_fs.open_output_stream(path)
        elif mode == "r":
            if self._sequential_read:  # pragma: no cover
                file = self._arrow_fs.open_input_stream(path)
            else:
                file = self._arrow_fs.open_input_file(path)
        elif mode == "a":
            file = self._arrow_fs.open_append_stream(path)
        else:  # pragma: no cover
            raise ValueError(
                f'mode can only be "wb", "rb" and "ab" for '
                f"arrow based filesystem, got {mode}"
            )

        _file_to_filesystems[file] = self._arrow_fs
        return file

    @implements(FileSystem.walk)
    def walk(self, path: path_type) -> Iterator[Tuple[str, List[str], List[str]]]:
        path = self._process_path(path)
        q = [path]
        while q:
            curr = q.pop(0)
            file_selector: FileSelector = FileSelector(curr)
            dirs, files = [], []
            for info in self._arrow_fs.get_file_info(file_selector):
                if info.type == FileType.File:
                    files.append(info.base_name)
                elif info.type == FileType.Directory:
                    dirs.append(info.base_name)
                    q.append(info.path)
                else:  # pragma: no cover
                    continue
            yield curr, dirs, files

    @implements(FileSystem.glob)
    def glob(self, path: path_type, recursive: bool = False) -> List[path_type]:
        from ._glob import FileSystemGlob

        path = self._process_path(path)
        return FileSystemGlob(self).glob(path, recursive=recursive)


class ArrowBasedLocalFileSystem(ArrowBasedFileSystem):
    def __init__(self):
        super().__init__(ArrowLocalFileSystem())

    _instance = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = ArrowBasedLocalFileSystem()
        return cls._instance


class HadoopFileSystem(ArrowBasedFileSystem):
    def __init__(
        self,
        host="default",
        port=0,
        user=None,
        kerb_ticket=None,
        driver="libhdfs",
        extra_conf=None,
    ):
        assert driver == "libhdfs"
        if "HADOOP_HOME" in os.environ and "CLASSPATH" not in os.environ:
            classpath_proc = subprocess.run(
                [os.environ["HADOOP_HOME"] + "/bin/hdfs", "classpath", "--glob"],
                stdout=subprocess.PIPE,
            )
            os.environ["CLASSPATH"] = classpath_proc.stdout.decode().strip()
        arrow_fs = ArrowHadoopFileSystem(
            host=host,
            port=port,
            user=user,
            kerb_ticket=kerb_ticket,
            extra_conf=extra_conf,
        )
        super().__init__(arrow_fs)

    @staticmethod
    def _process_path(path):
        path = ArrowBasedFileSystem._process_path(path)
        # use urlparse to extract path from like:
        # hdfs://localhost:8020/tmp/test/simple_test.csv,
        # due to the reason that pa.fs.HadoopFileSystem cannot accept
        # path with hdfs:// prefix
        if path.startswith("hdfs://"):
            return urlparse(path).path
        else:
            return path



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from numbers import Number
from typing import Dict, Type


class BaseActorDriver(ABC):
    @classmethod
    @abstractmethod
    def setup_cluster(cls, address_to_resources: Dict[str, Dict[str, Number]]):
        """
        Setup cluster according to given resources,
        resources is a dict, e.g. {'CPU': 3, 'GPU': 1}

        Parameters
        ----------
        address_to_resources: dict
            resources that required for each node.
        """
        pass


_backend_driver_cls: Dict[str, Type[BaseActorDriver]] = dict()


def register_backend_driver(scheme: str, cls: Type[BaseActorDriver]):
    assert issubclass(cls, BaseActorDriver)
    _backend_driver_cls[scheme] = cls



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...core import register_fetch_class, OutputType
from ...core.operand import Fetch, FetchShuffle, FetchMixin
from ...serialization.serializables import DataTypeField
from ..operands import TensorOperandMixin


class TensorFetchMixin(TensorOperandMixin, FetchMixin):
    __slots__ = ()
    _output_type_ = OutputType.tensor


class TensorFetch(TensorFetchMixin, Fetch):
    dtype = DataTypeField("dtype")

    def __init__(self, **kw):
        kw.pop("output_types", None)
        kw.pop("_output_types", None)
        super().__init__(**kw)

    def _new_chunks(self, inputs, kws=None, **kw):
        if "_key" in kw and self.source_key is None:
            self.source_key = kw["_key"]
        return super()._new_chunks(inputs, kws=kws, **kw)

    def _new_tileables(self, inputs, kws=None, **kw):
        if "_key" in kw and self.source_key is None:
            self.source_key = kw["_key"]
        return super()._new_tileables(inputs, kws=kws, **kw)


class TensorFetchShuffle(TensorFetchMixin, FetchShuffle):
    _dtype = DataTypeField("dtype")

    def __init__(self, **kw):
        kw.pop("output_types", None)
        kw.pop("_output_types", None)
        super().__init__(**kw)

    @property
    def dtype(self):
        return getattr(self, "_dtype", None)


register_fetch_class(OutputType.tensor, TensorFetch, TensorFetchShuffle)
register_fetch_class(OutputType.scalar, TensorFetch, TensorFetchShuffle)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import socket

import numpy as np

from ... import opcodes
from ...core.operand import OperandStage
from ...serialization.serializables import (
    FieldTypes,
    Int32Field,
    ListField,
    StringField,
)
from ...tensor.merge import TensorConcatenate
from ...utils import get_next_port
from ..operands import LearnOperand, LearnOperandMixin, OutputType


class CollectPorts(LearnOperand, LearnOperandMixin):
    _op_code_ = opcodes.COLLECT_PORTS

    _socket_type = Int32Field("socket_type")
    _index = Int32Field("index")
    _workers = ListField("workers", FieldTypes.string)
    _tileable_key = StringField("tileable_key")

    def __init__(
        self, workers=None, socket_type=None, tileable_key=None, index=None, **kw
    ):
        super().__init__(
            _socket_type=socket_type,
            _workers=workers,
            _tileable_key=tileable_key,
            _index=index,
            _pure_depends=[True],
            **kw
        )

    @property
    def socket_type(self):
        return self._socket_type

    @property
    def workers(self):
        return self._workers

    @property
    def tileable_key(self):
        return self._tileable_key

    def __call__(self, dep=None):
        self._output_types = [OutputType.tensor]
        if dep:
            deps = [dep]
        else:
            deps = None
        return self.new_tileable(deps, shape=(len(self.workers),), dtype=np.dtype(int))

    @classmethod
    def tile(cls, op: "CollectPorts"):
        chunks = []
        if op.inputs:
            chunk_iter = itertools.cycle(op.inputs[0].chunks)
        else:
            chunk_iter = itertools.repeat(None)
        for idx, (worker, inp) in enumerate(zip(op.workers, chunk_iter)):
            new_op = op.copy().reset_key()
            new_op._workers = [worker]
            new_op.expect_worker = worker
            new_op.stage = OperandStage.map
            new_op._tileable_key = op.outputs[0].key
            new_op._index = idx
            new_op._pure_depends = [True]
            inps = [inp] if inp else None
            chunks.append(
                new_op.new_chunk(inps, index=(idx,), shape=(1,), dtype=np.dtype(int))
            )

        concat_op = TensorConcatenate(axis=0, dtype=chunks[0].dtype)
        concat_chunk = concat_op.new_chunk(chunks, shape=(len(op.workers),), index=(0,))

        new_op = op.copy().reset_key()
        params = op.outputs[0].params
        params.update(dict(chunks=[concat_chunk], nsplits=((len(op.workers),),)))
        return new_op.new_tileables(op.inputs, **params)

    @classmethod
    def execute(cls, ctx, op):
        assert ctx.band[0] == op.expect_worker
        socket_type = op.socket_type or socket.SOCK_STREAM
        port_num = get_next_port(socket_type, occupy=False)
        ctx[op.outputs[0].key] = np.array([port_num], dtype=int)


def collect_ports(workers, input_tileable=None):
    op = CollectPorts(workers=workers)
    return op(input_tileable)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .... import oscar as mo
from ....utils import calc_size_by_str
from ...core import AbstractService
from .workerslot import WorkerSlotManagerActor
from .quota import WorkerQuotaManagerActor
from .execution import SubtaskExecutionActor, DEFAULT_SUBTASK_MAX_RETRIES


class SchedulingWorkerService(AbstractService):
    """
    Scheduling service on worker.

    Service Configuration
    ---------------------
    {
        "scheduling": {
            "mem_quota_size": "80%",
            "mem_hard_limit": "95%",
            "enable_kill_slot": true,
            "data_prepare_timeout": 600,
            "subtask_max_retries": 1
        }
    }
    """

    async def start(self):
        from .... import resource as mars_resource

        scheduling_config = self._config.get("scheduling", {})
        address = self._address

        total_mem = mars_resource.virtual_memory().total
        mem_quota_size = calc_size_by_str(
            scheduling_config.get("mem_quota_size", "80%"), total_mem
        )
        mem_hard_limit = calc_size_by_str(
            scheduling_config.get("mem_hard_limit", "95%"), total_mem
        )
        enable_kill_slot = scheduling_config.get("enable_kill_slot", True)
        subtask_max_retries = scheduling_config.get(
            "subtask_max_retries", DEFAULT_SUBTASK_MAX_RETRIES
        )
        data_prepare_timeout = scheduling_config.get("data_prepare_timeout", 600)

        await mo.create_actor(
            WorkerSlotManagerActor,
            uid=WorkerSlotManagerActor.default_uid(),
            address=address,
        )
        await mo.create_actor(
            WorkerQuotaManagerActor,
            default_config=dict(
                quota_size=mem_quota_size,
                hard_limit=mem_hard_limit,
                enable_kill_slot=enable_kill_slot,
            ),
            uid=WorkerQuotaManagerActor.default_uid(),
            address=address,
        )
        await mo.create_actor(
            SubtaskExecutionActor,
            subtask_max_retries=subtask_max_retries,
            enable_kill_slot=enable_kill_slot,
            data_prepare_timeout=data_prepare_timeout,
            uid=SubtaskExecutionActor.default_uid(),
            address=address,
        )

    async def stop(self):
        address = self._address

        await mo.destroy_actor(
            mo.create_actor_ref(
                uid=SubtaskExecutionActor.default_uid(), address=address
            )
        )
        await mo.destroy_actor(
            mo.create_actor_ref(
                uid=WorkerQuotaManagerActor.default_uid(), address=address
            )
        )
        await mo.destroy_actor(
            mo.create_actor_ref(
                uid=WorkerSlotManagerActor.default_uid(), address=address
            )
        )



# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import os
import signal
import uuid
from collections import defaultdict
from typing import AsyncGenerator, Dict, List, Optional, TypeVar

from ... import oscar as mo
from ...services import NodeRole
from ...services.cluster.backends import (
    register_cluster_backend,
    AbstractClusterBackend,
)
from ...utils import to_binary, to_str
from ..utils import wait_all_supervisors_ready
from .config import MarsSupervisorConfig, MarsWorkerConfig

try:
    from skein import (
        ApplicationClient,
        Client as SkeinClient,
        properties as skein_props,
        ConnectionError as SkeinConnectionError,
        SkeinError,
    )
except ImportError:  # pragma: no cover
    ApplicationClient, SkeinClient, skein_props = None, None, None
    SkeinConnectionError, SkeinError = None, None

RetType = TypeVar("RetType")
logger = logging.getLogger(__name__)

_role_to_config = {
    NodeRole.SUPERVISOR: MarsSupervisorConfig,
    NodeRole.WORKER: MarsWorkerConfig,
}


class YarnNodeWatchActor(mo.Actor):
    def __init__(self):
        assert ApplicationClient is not None
        self._app_client = ApplicationClient.from_current()

        self._nodes = defaultdict(set)
        self._supervisor_watch_task = None
        self._role_to_events = defaultdict(list)

    async def __post_create__(self):
        self._supervisor_watch_task = asyncio.create_task(
            self._watch_nodes(NodeRole.SUPERVISOR)
        )

    async def __pre_destroy__(self):
        if self._supervisor_watch_task is not None:  # pragma: no branch
            self._watch_task.cancel()

    async def get_container_mappings(self, role: NodeRole) -> Dict[str, str]:
        key_prefix = _role_to_config[role].service_name

        container_specs = await asyncio.to_thread(
            self._app_client.get_containers, [key_prefix]
        )
        cid_to_endpoint = {c.yarn_container_id: None for c in container_specs}

        prefixes = await asyncio.to_thread(self._app_client.kv.get_prefix, key_prefix)
        for val in prefixes.values():
            ep, cid = to_str(val).split("@", 1)
            cid_to_endpoint[cid] = ep
        return cid_to_endpoint

    async def _watch_nodes(self, role: NodeRole):
        while True:
            try:
                mappings = await self.get_container_mappings(role)
                eps = set(v for v in mappings.values() if v is not None)

                if eps != self._nodes[role]:
                    logger.info("New endpoints retrieved: %r", eps)
                    events = self._role_to_events.pop(role, [])
                    for ev in events:
                        ev.set()
                    self._nodes[role] = eps
                await asyncio.sleep(1)
            except SkeinConnectionError:  # pragma: no cover
                logger.warning("Skein application down, process will terminate")
                os.kill(os.getpid(), signal.SIGTERM)
            except (SkeinError, asyncio.CancelledError):  # pragma: no cover
                logger.exception("Error when watching nodes")
                break

    async def get_nodes(self, role: NodeRole) -> List[str]:
        if not self._nodes[role]:
            mappings = await self.get_container_mappings(role)
            eps = set(v for v in mappings.values() if v is not None)
            self._nodes[role] = eps
        return list(self._nodes[role])

    async def wait_nodes(self, role: NodeRole):
        event = asyncio.Event()
        self._role_to_events[role].append(event)

        async def waiter():
            await event.wait()
            return list(self._supervisors)

        return waiter()


@register_cluster_backend
class YarnClusterBackend(AbstractClusterBackend):
    name = "yarn"

    def __init__(self, pool_address: str, watch_ref: mo.ActorRef = None):
        self._pool_address = pool_address
        self._watch_ref = watch_ref

    @classmethod
    async def create(
        cls, node_role: NodeRole, lookup_address: Optional[str], pool_address: str
    ) -> "AbstractClusterBackend":
        try:
            ref = await mo.create_actor(
                YarnNodeWatchActor,
                uid=YarnNodeWatchActor.default_uid(),
                address=pool_address,
            )
        except mo.ActorAlreadyExist:  # pragma: no cover
            ref = await mo.actor_ref(
                YarnNodeWatchActor.default_uid(), address=pool_address
            )
        return YarnClusterBackend(pool_address, ref)

    async def get_supervisors(self, filter_ready: bool = True) -> List[str]:
        if filter_ready:
            return await self._watch_ref.get_nodes(NodeRole.SUPERVISOR)
        else:
            mapping = await self._watch_ref.get_container_mappings(NodeRole.SUPERVISOR)
            return [v if v is not None else k for k, v in mapping.items()]

    async def watch_supervisors(self) -> AsyncGenerator[List[str], None]:
        while True:
            yield await self._watch_ref.wait_nodes(NodeRole.SUPERVISOR)

    async def request_worker(
        self, worker_cpu: int = None, worker_mem: int = None, timeout: int = None
    ) -> str:
        raise NotImplementedError

    async def release_worker(self, address: str):
        raise NotImplementedError

    async def reconstruct_worker(self, address: str):
        raise NotImplementedError


class YarnServiceMixin(object):
    service_name = None

    @property
    def app_client(self):
        if not hasattr(self, "_app_client"):
            self._app_client = ApplicationClient.from_current()
        return self._app_client

    def get_container_ip(self):
        svc_containers = self.app_client.get_containers([self.service_name])
        container = next(
            c
            for c in svc_containers
            if c.yarn_container_id == skein_props["yarn_container_id"]
        )
        return container.yarn_node_http_address.split(":")[0]

    def register_endpoint(self, prefix: str = None, endpoint: str = None):
        prefix = prefix or self.service_name
        endpoint = endpoint or self.args.endpoint

        container_key = prefix + "-" + str(uuid.uuid1())
        self.app_client.kv[container_key] = to_binary(
            f'{endpoint}@{skein_props["yarn_container_id"]}'
        )

    async def wait_all_supervisors_ready(self):
        """
        Wait till all containers are ready, both in yarn and in Cluster Service
        """
        await wait_all_supervisors_ready(self.args.endpoint)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import Int64Field
from ..utils import gen_random_seeds
from .core import TensorRandomOperandMixin, TensorSimpleRandomData


class TensorRandomIntegers(TensorSimpleRandomData, TensorRandomOperandMixin):
    _op_type_ = OperandDef.RAND_RANDOM_INTEGERS

    _fields_ = "low", "high", "size"
    low = Int64Field("low")
    high = Int64Field("high")
    _func_name = "random_integers"

    def __call__(self, chunk_size=None):
        return self.new_tensor(None, None, raw_chunk_size=chunk_size)


def random_integers(random_state, low, high=None, size=None, chunk_size=None, gpu=None):
    """
    Random integers of type mt.int between `low` and `high`, inclusive.

    Return random integers of type mt.int from the "discrete uniform"
    distribution in the closed interval [`low`, `high`].  If `high` is
    None (the default), then results are from [1, `low`]. The np.int
    type translates to the C long type used by Python 2 for "short"
    integers and its precision is platform dependent.

    This function has been deprecated. Use randint instead.

    Parameters
    ----------
    low : int
        Lowest (signed) integer to be drawn from the distribution (unless
        ``high=None``, in which case this parameter is the *highest* such
        integer).
    high : int, optional
        If provided, the largest (signed) integer to be drawn from the
        distribution (see above for behavior if ``high=None``).
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  Default is None, in which case a
        single value is returned.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : int or Tensor of ints
        `size`-shaped array of random integers from the appropriate
        distribution, or a single such random int if `size` not provided.

    See Also
    --------
    random.randint : Similar to `random_integers`, only for the half-open
        interval [`low`, `high`), and 0 is the lowest value if `high` is
        omitted.

    Notes
    -----
    To sample from N evenly spaced floating-point numbers between a and b,
    use::

      a + (b - a) * (np.random.random_integers(N) - 1) / (N - 1.)

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.random.random_integers(5).execute()
    4
    >>> type(mt.random.random_integers(5).execute())
    <type 'int'>
    >>> mt.random.random_integers(5, size=(3,2)).execute()
    array([[5, 4],
           [3, 3],
           [4, 5]])

    Choose five random numbers from the set of five evenly-spaced
    numbers between 0 and 2.5, inclusive (*i.e.*, from the set
    :math:`{0, 5/8, 10/8, 15/8, 20/8}`):

    >>> (2.5 * (mt.random.random_integers(5, size=(5,)) - 1) / 4.).execute()
    array([ 0.625,  1.25 ,  0.625,  0.625,  2.5  ])

    Roll two six sided dice 1000 times and sum the results:

    >>> d1 = mt.random.random_integers(1, 6, 1000)
    >>> d2 = mt.random.random_integers(1, 6, 1000)
    >>> dsums = d1 + d2

    Display results as a histogram:

    >>> import matplotlib.pyplot as plt
    >>> count, bins, ignored = plt.hist(dsums.execute(), 11, normed=True)
    >>> plt.show()
    """
    size = random_state._handle_size(size)
    seed = gen_random_seeds(1, random_state.to_numpy())[0]
    op = TensorRandomIntegers(
        seed=seed, size=size, dtype=np.dtype(int), low=low, high=high, gpu=gpu
    )
    return op(chunk_size=chunk_size)



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from .... import oscar as mo
from ....lib.aio import alru_cache
from ...cluster import ClusterAPI
from ..store import get_meta_store


class WorkerMetaStoreManagerActor(mo.Actor):
    def __init__(self, meta_store_name: str, config: Dict):
        self._meta_store_name = meta_store_name
        self._meta_store_type = get_meta_store(meta_store_name)
        self._config = config
        self._meta_init_kwargs = None

        self._cluster_api = None

    async def __post_create__(self):
        self._meta_init_kwargs = await self._meta_store_type.create(self._config)
        self._cluster_api = await ClusterAPI.create(self.address)

    @alru_cache(cache_exceptions=False)
    async def _get_supervisor_address(self, session_id: str):
        [address] = await self._cluster_api.get_supervisors_by_keys([session_id])
        return address

    async def new_session_meta_store(self, session_id: str) -> mo.ActorRef:
        from ..supervisor.core import MetaStoreActor

        try:
            ref = await mo.create_actor(
                WorkerMetaStoreActor,
                self._meta_store_name,
                session_id,
                uid=WorkerMetaStoreActor.gen_uid(session_id),
                address=self.address,
                **self._meta_init_kwargs,
            )
            supervisor_address = await self._get_supervisor_address(session_id)
            supervisor_meta_store_ref = await mo.actor_ref(
                uid=MetaStoreActor.gen_uid(session_id), address=supervisor_address
            )
            # register worker meta store,
            # when session destroyed, this worker meta store actor will be removed
            await supervisor_meta_store_ref.add_worker_meta_store(ref)
        except mo.ActorAlreadyExist:
            ref = await mo.actor_ref(
                uid=WorkerMetaStoreActor.gen_uid(session_id), address=self.address
            )
        return ref


class WorkerMetaStoreActor(mo.Actor):
    def __init__(self, meta_store_name: str, session_id: str, **meta_store_kwargs):
        meta_store_type = get_meta_store(meta_store_name)
        self._store = meta_store_type(session_id, **meta_store_kwargs)

    @staticmethod
    def gen_uid(session_id: str):
        return f"{session_id}_worker_meta"

    def __getattr__(self, attr):
        return getattr(self._store, attr)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...config import options
from .core import DataFrameCumReductionOperand, DataFrameCumReductionMixin


class DataFrameCummax(DataFrameCumReductionOperand, DataFrameCumReductionMixin):
    _op_type_ = OperandDef.CUMMAX
    _func_name = "cummax"


def cummax(df, axis=None, skipna=True):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameCummax(
        axis=axis,
        skipna=skipna,
        output_types=df.op.output_types,
        use_inf_as_na=use_inf_as_na,
    )
    return op(df)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List

try:
    import scipy.sparse as sps
except ImportError:  # pragma: no cover
    sps = None

from ..lib.sparse import SparseNDArray
from .core import Serializer, serialize, deserialize, buffered


class SparseNDArraySerializer(Serializer):
    @buffered
    def serial(self, obj: Any, context: Dict):
        raw_header, raw_buffers = serialize(obj.raw, context)
        return (raw_header, obj.shape), raw_buffers, True

    def deserial(self, serialized: Dict, context: Dict, subs: List):
        raw_header, obj_shape = serialized
        raw_csr = deserialize(raw_header, subs)
        return SparseNDArray(raw_csr, shape=tuple(obj_shape))


if sps:  # pragma: no branch
    SparseNDArraySerializer.register(SparseNDArray)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
from pandas.api.types import is_list_like

from ... import opcodes
from ...core import OutputType, recursive_tile
from ...serialization.serializables import KeyField, AnyField
from ...tensor.core import TENSOR_TYPE
from ...utils import pd_release_version
from ..core import DATAFRAME_TYPE, SERIES_TYPE, DataFrame
from ..initializer import DataFrame as asframe, Series as asseries
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index

# in pandas 1.0.x, __setitem__ with a list with missing items are not allowed
_allow_set_missing_list = pd_release_version[:2] >= (1, 1)


class DataFrameSetitem(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.INDEXSETVALUE

    _target = KeyField("target")
    _indexes = AnyField("indexes")
    _value = AnyField("value")

    def __init__(self, target=None, indexes=None, value=None, output_types=None, **kw):
        super().__init__(
            _target=target,
            _indexes=indexes,
            _value=value,
            _output_types=output_types,
            **kw
        )
        if self.output_types is None:
            self.output_types = [OutputType.dataframe]

    @property
    def target(self):
        return self._target

    @property
    def indexes(self):
        return self._indexes

    @property
    def value(self):
        return self._value

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._target = self._inputs[0]
        if len(inputs) > 1:
            self._value = self._inputs[-1]

    @staticmethod
    def _is_scalar_tensor(t):
        return isinstance(t, TENSOR_TYPE) and t.ndim == 0

    def __call__(self, target: DataFrame, value):
        raw_target = target

        inputs = [target]
        if np.isscalar(value):
            value_dtype = np.array(value).dtype
        elif self._is_scalar_tensor(value):
            inputs.append(value)
            value_dtype = value.dtype
        else:
            if isinstance(value, (pd.Series, SERIES_TYPE)):
                value = asseries(value)
                value_dtype = value.dtype
            elif isinstance(value, (pd.DataFrame, DATAFRAME_TYPE)):
                if len(self.indexes) != value.shape[1]:  # pragma: no cover
                    raise ValueError("Columns must be same length as key")

                value = asframe(value)
                value_dtype = pd.Series(list(value.dtypes), index=self._indexes)
            elif is_list_like(value) or isinstance(value, TENSOR_TYPE):
                # convert to numpy to get actual dim and shape
                if is_list_like(value):
                    value = np.array(value)

                if value.ndim == 1:
                    value = asseries(value, index=target.index)
                    value_dtype = value.dtype
                else:
                    if len(self.indexes) != value.shape[1]:  # pragma: no cover
                        raise ValueError("Columns must be same length as key")

                    value = asframe(value, index=target.index)
                    value_dtype = pd.Series(list(value.dtypes), index=self._indexes)
            else:  # pragma: no cover
                raise TypeError(
                    "Wrong value type, could be one of scalar, Series or tensor"
                )

            if target.shape[0] == 0:
                # target empty, reindex target first
                target = target.reindex(value.index)
                inputs[0] = target
            elif value.index_value.key != target.index_value.key:
                # need reindex when target df is not empty and index different
                value = value.reindex(target.index)
            inputs.append(value)

        index_value = target.index_value
        dtypes = target.dtypes.copy(deep=True)

        try:
            dtypes.loc[self._indexes] = value_dtype
        except KeyError:
            # when some index not exist, try update one by one
            if isinstance(value_dtype, pd.Series):
                for idx in self._indexes:
                    dtypes.loc[idx] = value_dtype.loc[idx]
            else:
                for idx in self._indexes:
                    dtypes.loc[idx] = value_dtype

        columns_value = parse_index(dtypes.index, store_data=True)
        ret = self.new_dataframe(
            inputs,
            shape=(target.shape[0], len(dtypes)),
            dtypes=dtypes,
            index_value=index_value,
            columns_value=columns_value,
        )
        raw_target.data = ret.data

    @classmethod
    def tile(cls, op: "DataFrameSetitem"):
        out = op.outputs[0]
        target = op.target
        value = op.value
        indexes = op.indexes
        columns = target.columns_value.to_pandas()
        is_value_scalar = np.isscalar(value) or cls._is_scalar_tensor(value)
        has_multiple_cols = getattr(out.dtypes[indexes], "ndim", 0) > 0
        target_index_to_value = dict()

        if has_multiple_cols:
            append_cols = [c for c in indexes if c not in columns]
        else:
            append_cols = [indexes] if indexes not in columns else []

        if not is_value_scalar:
            rechunk_arg = {}

            # check if all chunk's index_value are identical
            target_chunk_index_values = [
                c.index_value for c in target.chunks if c.index[1] == 0
            ]
            value_chunk_index_values = [v.index_value for v in value.chunks]
            is_identical = len(target_chunk_index_values) == len(
                value_chunk_index_values
            ) and all(
                c.key == v.key
                for c, v in zip(target_chunk_index_values, value_chunk_index_values)
            )
            if not is_identical:
                # do rechunk
                if any(np.isnan(s) for s in target.nsplits[0]) or any(
                    np.isnan(s) for s in value.nsplits[0]
                ):  # pragma: no cover
                    yield

                rechunk_arg[0] = target.nsplits[0]

            if isinstance(value, DATAFRAME_TYPE):
                # rechunk in column dim given distribution of indexes in target chunks
                col_nsplits = []
                for head_chunk in target.cix[0, :]:
                    new_indexes = [vc for vc in indexes if vc in head_chunk.dtypes]
                    if new_indexes:
                        target_index_to_value[head_chunk.index[1]] = len(col_nsplits)
                        col_nsplits.append(len(new_indexes))

                if not col_nsplits:
                    # purely new chunks, just update last column of chunks
                    target_index_to_value[target.chunk_shape[1] - 1] = 0
                    col_nsplits = [len(append_cols)]
                else:
                    col_nsplits[-1] += len(append_cols)
                rechunk_arg[1] = col_nsplits

            if rechunk_arg:
                value = yield from recursive_tile(value.rechunk(rechunk_arg))

        out_chunks = []
        nsplits = [list(ns) for ns in target.nsplits]

        nsplits[1][-1] += len(append_cols)

        column_chunk_shape = target.chunk_shape[1]
        for c in target.chunks:
            result_chunk = c

            if has_multiple_cols:
                new_indexes = [vc for vc in indexes if vc in c.dtypes]
            else:
                new_indexes = [indexes] if indexes in c.dtypes else []

            if c.index[-1] == column_chunk_shape - 1:
                new_indexes.extend(append_cols)

            if new_indexes:
                # update needed on current chunk
                chunk_op = op.copy().reset_key()
                chunk_op._indexes = new_indexes if has_multiple_cols else new_indexes[0]

                if pd.api.types.is_scalar(value):
                    chunk_inputs = [c]
                elif is_value_scalar:
                    chunk_inputs = [c, value.chunks[0]]
                else:
                    # get proper chunk from value chunks
                    if has_multiple_cols:
                        value_chunk = value.cix[
                            c.index[0], target_index_to_value[c.index[1]]
                        ]
                    else:
                        value_chunk = value.cix[
                            c.index[0],
                        ]

                    chunk_inputs = [c, value_chunk]

                dtypes, shape, columns_value = c.dtypes, c.shape, c.columns_value

                if append_cols and c.index[-1] == column_chunk_shape - 1:
                    # some columns appended at the last column of chunks
                    shape = (shape[0], shape[1] + len(append_cols))
                    dtypes = pd.concat([dtypes, out.dtypes.iloc[-len(append_cols) :]])
                    columns_value = parse_index(dtypes.index, store_data=True)

                result_chunk = chunk_op.new_chunk(
                    chunk_inputs,
                    shape=shape,
                    dtypes=dtypes,
                    index_value=c.index_value,
                    columns_value=columns_value,
                    index=c.index,
                )
            out_chunks.append(result_chunk)

        params = out.params
        params["nsplits"] = tuple(tuple(ns) for ns in nsplits)
        params["chunks"] = out_chunks
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def estimate_size(cls, ctx: dict, op: "DataFrameSetitem"):
        result_size = ctx[op.target.key][0]
        ctx[op.outputs[0].key] = (result_size, result_size)

    @classmethod
    def execute(cls, ctx, op: "DataFrameSetitem"):
        target = ctx[op.target.key].copy()
        value = ctx[op.value.key] if not np.isscalar(op.value) else op.value
        try:

            target[op.indexes] = value
        except KeyError:
            if _allow_set_missing_list:  # pragma: no cover
                raise
            else:
                existing = set(target.columns)
                new_columns = target.columns.append(
                    pd.Index([idx for idx in op.indexes if idx not in existing])
                )
                target = target.reindex(new_columns, axis=1)
                target[op.indexes] = value

        ctx[op.outputs[0].key] = target


def dataframe_setitem(df, col, value):
    op = DataFrameSetitem(target=df, indexes=col, value=value)
    return op(df, value)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import AnyField
from ...utils import has_unknown_shape
from ..core import Tensor
from ..datasource import tensor as astensor
from ..operands import TensorOperand, TensorOperandMixin
from ..utils import calc_sliced_size
from .core import chunk_size_type, get_nsplits, gen_rechunk_infos


class TensorRechunk(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.RECHUNK

    chunk_size = AnyField("chunk_size")

    def __call__(self, tensor: Tensor):
        return self.new_tensor([tensor], tensor.shape, order=tensor.order)

    @classmethod
    def tile(cls, op: "TensorRechunk"):
        from ..indexing.slice import TensorSlice
        from ..merge.concatenate import TensorConcatenate

        if has_unknown_shape(*op.inputs):
            yield

        out = op.outputs[0]
        tensor = astensor(op.inputs[0])
        chunk_size = get_nsplits(tensor, op.chunk_size, tensor.dtype.itemsize)
        if chunk_size == tensor.nsplits:
            return [tensor]

        rechunk_infos = gen_rechunk_infos(tensor, chunk_size)
        out_chunks = []
        for rechunk_info in rechunk_infos:
            chunk_index = rechunk_info.out_index
            shape = rechunk_info.shape
            inp_chunks = rechunk_info.input_chunks
            inp_chunk_slices = rechunk_info.input_slices
            inp_slice_chunks = []
            for inp_chunk, inp_chunk_slice in zip(inp_chunks, inp_chunk_slices):
                if all(slc == slice(None) for slc in inp_chunk_slice):
                    inp_slice_chunks.append(inp_chunk)
                else:
                    slc_chunk = TensorSlice(slices=list(inp_chunk_slice)).new_chunk(
                        [inp_chunk],
                        dtype=inp_chunk.dtype,
                        shape=tuple(
                            calc_sliced_size(s, slc)
                            for s, slc in zip(inp_chunk.shape, inp_chunk_slice)
                        ),
                        index=inp_chunk.index,
                    )
                    inp_slice_chunks.append(slc_chunk)

            if len(inp_slice_chunks) > 1 or inp_slice_chunks[0].index != chunk_index:
                chunk_op = TensorConcatenate()
                out_chunk = chunk_op.new_chunk(
                    inp_slice_chunks,
                    shape=shape,
                    index=chunk_index,
                    dtype=out.dtype,
                    order=out.order,
                )
                out_chunks.append(out_chunk)
            else:
                out_chunks.append(inp_slice_chunks[0])

        new_op = op.copy()
        params = out.params
        params["nsplits"] = chunk_size
        params["chunks"] = out_chunks
        tensor = new_op.new_tileable(op.inputs, kws=[params])

        if op.reassign_worker:
            for c in tensor.chunks:
                c.op.reassign_worker = True

        return [tensor]


def rechunk(
    tensor: Tensor, chunk_size: chunk_size_type, reassign_worker=False
) -> Tensor:
    if not any(np.isnan(s) for s in tensor.shape) and not tensor.is_coarse():
        if not has_unknown_shape(tensor):
            # do client check only when tensor has no unknown shape,
            # otherwise, recalculate chunk_size in `tile`
            chunk_size = get_nsplits(tensor, chunk_size, tensor.dtype.itemsize)
            if chunk_size == tensor.nsplits:
                return tensor

    op = TensorRechunk(
        chunk_size=chunk_size,
        reassign_worker=reassign_worker,
        dtype=tensor.dtype,
        sparse=tensor.issparse(),
    )
    return op(tensor)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

from ... import opcodes as OperandDef
from ...core import recursive_tile
from ...lib.filesystem import open_file, get_fs
from ...serialization.serializables import (
    KeyField,
    AnyField,
    StringField,
    ListField,
    BoolField,
    DictField,
)
from ...utils import has_unknown_shape
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index
from ..datasource.read_parquet import check_engine

try:
    import pyarrow.parquet as pq
    import pyarrow as pa
except ImportError:  # pragma: no cover
    pq = None
    pa = None


class DataFrameToParquet(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.TO_PARQUET

    _input = KeyField("input")
    _path = AnyField("path")
    _engine = StringField("engine")
    _index = BoolField("index")
    _compression = AnyField("compression")
    _partition_cols = ListField("partition_cols")
    _additional_kwargs = DictField("additional_kwargs")
    _storage_options = DictField("storage_options")

    def __init__(
        self,
        path=None,
        engine=None,
        index=None,
        compression=None,
        partition_cols=None,
        storage_options=None,
        additional_kwargs=None,
        **kw,
    ):
        super().__init__(
            _path=path,
            _engine=engine,
            _index=index,
            _compression=compression,
            _partition_cols=partition_cols,
            _storage_options=storage_options,
            _additional_kwargs=additional_kwargs,
            **kw,
        )

    @property
    def input(self):
        return self._input

    @property
    def path(self):
        return self._path

    @property
    def engine(self):
        return self._engine

    @property
    def index(self):
        return self._index

    @property
    def compression(self):
        return self._compression

    @property
    def partition_cols(self):
        return self._partition_cols

    @property
    def storage_options(self):
        return self._storage_options

    @property
    def additional_kwargs(self):
        return self._additional_kwargs

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    @classmethod
    def _get_path(cls, path, i):
        if "*" not in path:
            return path
        return path.replace("*", str(i))

    @classmethod
    def tile(cls, op):
        in_df = op.input
        out_df = op.outputs[0]

        # make sure only 1 chunk on the column axis
        if in_df.chunk_shape[1] > 1:
            if has_unknown_shape(in_df):
                yield
            in_df = yield from recursive_tile(in_df.rechunk({1: in_df.shape[1]}))

        out_chunks = []
        for chunk in in_df.chunks:
            chunk_op = op.copy().reset_key()
            index_value = parse_index(chunk.index_value.to_pandas()[:0], chunk)
            out_chunk = chunk_op.new_chunk(
                [chunk],
                shape=(0, 0),
                index_value=index_value,
                columns_value=out_df.columns_value,
                dtypes=out_df.dtypes,
                index=chunk.index,
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = out_df.params.copy()
        params.update(
            dict(chunks=out_chunks, nsplits=((0,) * in_df.chunk_shape[0], (0,)))
        )
        return new_op.new_tileables([in_df], **params)

    @classmethod
    def execute(cls, ctx, op):
        df = ctx[op.input.key]
        out = op.outputs[0]
        i = op.outputs[0].index[0]
        path = op.path
        has_wildcard = False
        if "*" in path:
            path = path.replace("*", str(i))
            has_wildcard = True

        if op.partition_cols is None:
            if not has_wildcard:
                fs = get_fs(path, op.storage_options)
                path = fs.pathsep.join([path.rstrip(fs.pathsep), f"{i}.parquet"])
            if op.engine == "fastparquet":
                df.to_parquet(
                    path,
                    engine=op.engine,
                    compression=op.compression,
                    index=op.index,
                    open_with=open_file,
                    **op.additional_kwargs,
                )
            else:
                with open_file(
                    path, mode="wb", storage_options=op.storage_options
                ) as f:
                    df.to_parquet(
                        f,
                        engine=op.engine,
                        compression=op.compression,
                        index=op.index,
                        **op.additional_kwargs or dict(),
                    )
        else:
            if op.engine == "pyarrow":
                pq.write_to_dataset(
                    pa.Table.from_pandas(df), path, partition_cols=op.partition_cols
                )
            else:  # pragma: no cover
                raise NotImplementedError(
                    "Only support pyarrow engine when specify `partition_cols`."
                )

        ctx[out.key] = pd.DataFrame()

    def __call__(self, df):
        index_value = parse_index(df.index_value.to_pandas()[:0], df)
        columns_value = parse_index(df.columns_value.to_pandas()[:0], store_data=True)
        return self.new_dataframe(
            [df],
            shape=(0, 0),
            dtypes=df.dtypes[:0],
            index_value=index_value,
            columns_value=columns_value,
        )


def to_parquet(
    df,
    path,
    engine="auto",
    compression="snappy",
    index=None,
    partition_cols=None,
    **kwargs,
):
    """
    Write a DataFrame to the binary parquet format, each chunk will be
    written to a Parquet file.

    Parameters
    ----------
    path : str or file-like object
        If path is a string with wildcard e.g. '/to/path/out-*.parquet',
        `to_parquet` will try to write multiple files, for instance,
        chunk (0, 0) will write data into '/to/path/out-0.parquet'.
        If path is a string without wildcard, we will treat it as a directory.

    engine : {'auto', 'pyarrow', 'fastparquet'}, default 'auto'
        Parquet library to use. The default behavior is to try 'pyarrow',
        falling back to 'fastparquet' if 'pyarrow' is unavailable.

    compression : {'snappy', 'gzip', 'brotli', None}, default 'snappy'
        Name of the compression to use. Use ``None`` for no compression.

    index : bool, default None
        If ``True``, include the dataframe's index(es) in the file output.
        If ``False``, they will not be written to the file.
        If ``None``, similar to ``True`` the dataframe's index(es)
        will be saved. However, instead of being saved as values,
        the RangeIndex will be stored as a range in the metadata so it
        doesn't require much space and is faster. Other indexes will
        be included as columns in the file output.

    partition_cols : list, optional, default None
        Column names by which to partition the dataset.
        Columns are partitioned in the order they are given.
        Must be None if path is not a string.

    **kwargs
        Additional arguments passed to the parquet library.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame(data={'col1': [1, 2], 'col2': [3, 4]})
    >>> df.to_parquet('*.parquet.gzip',
    ...               compression='gzip').execute()  # doctest: +SKIP
    >>> md.read_parquet('*.parquet.gzip').execute()  # doctest: +SKIP
       col1  col2
    0     1     3
    1     2     4

    >>> import io
    >>> f = io.BytesIO()
    >>> df.to_parquet(f).execute()
    >>> f.seek(0)
    0
    >>> content = f.read()
    """
    engine = check_engine(engine)
    op = DataFrameToParquet(
        path=path,
        engine=engine,
        compression=compression,
        index=index,
        partition_cols=partition_cols,
        additional_kwargs=kwargs,
    )
    return op(df)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numbers
from abc import ABCMeta, abstractmethod
from itertools import chain
from math import ceil, floor

import numpy as np

from ...core import ExecutableTuple
from ... import tensor as mt
from ...tensor.utils import check_random_state
from ..utils import shuffle as shuffle_arrays
from ..utils.validation import indexable, _num_samples


def train_test_split(*arrays, **options):
    """Split arrays or matrices into random train and test subsets

    Parameters
    ----------
    *arrays : sequence of indexables with same length / shape[0]
        Allowed inputs are lists, numpy arrays, scipy-sparse
        matrices or pandas dataframes.

    test_size : float, int or None, optional (default=None)
        If float, should be between 0.0 and 1.0 and represent the proportion
        of the dataset to include in the test split. If int, represents the
        absolute number of test samples. If None, the value is set to the
        complement of the train size. If ``train_size`` is also None, it will
        be set to 0.25.

    train_size : float, int, or None, (default=None)
        If float, should be between 0.0 and 1.0 and represent the
        proportion of the dataset to include in the train split. If
        int, represents the absolute number of train samples. If None,
        the value is automatically set to the complement of the test size.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    shuffle : boolean, optional (default=True)
        Whether or not to shuffle the data before splitting. If shuffle=False
        then stratify must be None.

    stratify : array-like or None (default=None)
        If not None, data is split in a stratified fashion, using this as
        the class labels.

    Returns
    -------
    splitting : list, length=2 * len(arrays)
        List containing train-test split of inputs.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.learn.model_selection import train_test_split
    >>> X, y = mt.arange(10).reshape((5, 2)), range(5)
    >>> X.execute()
    array([[0, 1],
           [2, 3],
           [4, 5],
           [6, 7],
           [8, 9]])
    >>> list(y)
    [0, 1, 2, 3, 4]

    >>> X_train, X_test, y_train, y_test = train_test_split(
    ...     X, y, test_size=0.33, random_state=42)
    ...
    >>> X_train.execute()
    array([[8, 9],
           [0, 1],
           [4, 5]])
    >>> y_train.execute()
    array([4, 0, 2])
    >>> X_test.execute()
    array([[2, 3],
           [6, 7]])
    >>> y_test.execute()
    array([1, 3])

    >>> train_test_split(y, shuffle=False)
    [array([0, 1, 2]), array([3, 4])]

    """

    n_arrays = len(arrays)
    if n_arrays == 0:
        raise ValueError("At least one array required as input")
    test_size = options.pop("test_size", None)
    train_size = options.pop("train_size", None)
    random_state = options.pop("random_state", None)
    stratify = options.pop("stratify", None)
    shuffle = options.pop("shuffle", True)
    session = options.pop("session", None)
    run_kwargs = options.pop("run_kwargs", None)

    if options:
        raise TypeError(f"Invalid parameters passed: {options}")

    arrays = indexable(*arrays, session=session, run_kwargs=run_kwargs)

    n_samples = _num_samples(arrays[0])
    n_train, n_test = _validate_shuffle_split(
        n_samples, test_size, train_size, default_test_size=0.25
    )

    if shuffle is False:
        if stratify is not None:  # pragma: no cover
            raise ValueError(
                "Stratified train/test split is not implemented for shuffle=False"
            )

        iterables = ((a[:n_train], a[n_train : n_train + n_test]) for a in arrays)
    else:
        if stratify is not None:  # pragma: no cover
            raise NotImplementedError("stratify is not implemented yet")
        else:
            shuffled_arrays = shuffle_arrays(*arrays, random_state=random_state)
            if not isinstance(shuffled_arrays, tuple):
                shuffled_arrays = (shuffled_arrays,)
            iterables = (
                (a[:n_train], a[n_train : n_train + n_test]) for a in shuffled_arrays
            )

    return list(
        ExecutableTuple(chain.from_iterable(iterables)).execute(
            session=session, **(run_kwargs or dict())
        )
    )


def _validate_shuffle_split(n_samples, test_size, train_size, default_test_size=None):
    """
    Validation helper to check if the test/test sizes are meaningful wrt to the
    size of the data (n_samples)
    """
    if test_size is None and train_size is None:
        test_size = default_test_size

    test_size_type = np.asarray(test_size).dtype.kind
    train_size_type = np.asarray(train_size).dtype.kind

    if (
        test_size_type == "i"
        and (test_size >= n_samples or test_size <= 0)
        or test_size_type == "f"
        and (test_size <= 0 or test_size >= 1)
    ):
        raise ValueError(
            f"test_size={test_size} should be either positive and smaller"
            f" than the number of samples {n_samples} or a float in the "
            "(0, 1) range"
        )

    if (
        train_size_type == "i"
        and (train_size >= n_samples or train_size <= 0)
        or train_size_type == "f"
        and (train_size <= 0 or train_size >= 1)
    ):
        raise ValueError(
            f"train_size={train_size} should be either positive and smaller"
            f" than the number of samples {n_samples} or a float in the "
            "(0, 1) range"
        )

    if train_size is not None and train_size_type not in ("i", "f"):  # pragma: no cover
        raise ValueError(f"Invalid value for train_size: {train_size}")
    if test_size is not None and test_size_type not in ("i", "f"):
        raise ValueError(f"Invalid value for test_size: {test_size}")

    if train_size_type == "f" and test_size_type == "f" and train_size + test_size > 1:
        raise ValueError(
            f"The sum of test_size and train_size = {train_size + test_size}, "
            "should be in the (0, 1) range. Reduce test_size and/or train_size."
        )

    if test_size_type == "f":
        n_test = ceil(test_size * n_samples)
    elif test_size_type == "i":
        n_test = float(test_size)

    if train_size_type == "f":
        n_train = floor(train_size * n_samples)
    elif train_size_type == "i":  # pragma: no cover
        n_train = float(train_size)

    if train_size is None:
        n_train = n_samples - n_test
    elif test_size is None:
        n_test = n_samples - n_train

    if n_train + n_test > n_samples:  # pragma: no cover
        raise ValueError(
            f"The sum of train_size and test_size = {n_train + n_test}, "
            f"should be smaller than the number of samples {n_samples}. "
            "Reduce test_size and/or train_size."
        )

    n_train, n_test = int(n_train), int(n_test)

    if n_train == 0:  # pragma: no cover
        raise ValueError(
            f"With n_samples={n_samples}, test_size={test_size} and "
            f"train_size={train_size}, the resulting train set will "
            f"be empty. Adjust any of the aforementioned parameters."
        )

    return n_train, n_test


class BaseCrossValidator(metaclass=ABCMeta):
    """Base class for all cross-validators

    Implementations must define `_iter_test_masks` or `_iter_test_indices`.
    """

    def split(self, X, y=None, groups=None):  # pragma: no cover
        """Generate indices to split data into training and test set.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        y : array-like of shape (n_samples,)
            The target variable for supervised learning problems.

        groups : array-like of shape (n_samples,), default=None
            Group labels for the samples used while splitting the dataset into
            train/test set.

        Yields
        ------
        train : ndarray
            The training set indices for that split.

        test : ndarray
            The testing set indices for that split.
        """
        X, y, groups = indexable(X, y, groups)
        indices = mt.arange(_num_samples(X))
        for test_index in self._iter_test_masks(X, y, groups):
            train_index = indices[mt.logical_not(test_index)]
            test_index = indices[test_index]
            yield train_index, test_index

    # Since subclasses must implement either _iter_test_masks or
    # _iter_test_indices, neither can be abstract.
    def _iter_test_masks(self, X=None, y=None, groups=None):  # pragma: no cover
        """Generates boolean masks corresponding to test sets.

        By default, delegates to _iter_test_indices(X, y, groups)
        """
        for test_index in self._iter_test_indices(X, y, groups):
            test_mask = mt.zeros(_num_samples(X), dtype=bool)
            test_mask[test_index] = True
            yield test_mask

    def _iter_test_indices(self, X=None, y=None, groups=None):  # pragma: no cover
        """Generates integer indices corresponding to test sets."""
        raise NotImplementedError

    @abstractmethod
    def get_n_splits(self, X=None, y=None, groups=None):
        """Returns the number of splitting iterations in the cross-validator"""


class _BaseKFold(BaseCrossValidator, metaclass=ABCMeta):
    """Base class for KFold, GroupKFold, and StratifiedKFold"""

    @abstractmethod
    def __init__(self, n_splits, *, shuffle, random_state):
        if not isinstance(n_splits, numbers.Integral):
            raise ValueError(
                "The number of folds must be of Integral type. "
                "%s of type %s was passed." % (n_splits, type(n_splits))
            )
        n_splits = int(n_splits)

        if n_splits <= 1:
            raise ValueError(
                "k-fold cross-validation requires at least one"
                " train/test split by setting n_splits=2 or more,"
                " got n_splits={0}.".format(n_splits)
            )

        if not isinstance(shuffle, bool):
            raise TypeError("shuffle must be True or False; got {0}".format(shuffle))

        if not shuffle and random_state is not None:  # None is the default
            raise ValueError(
                "Setting a random_state has no effect since shuffle is "
                "False. You should leave "
                "random_state to its default (None), or set shuffle=True.",
            )

        self.n_splits = n_splits
        self.shuffle = shuffle
        self.random_state = random_state

    def get_n_splits(self, X=None, y=None, groups=None):
        """Returns the number of splitting iterations in the cross-validator

        Parameters
        ----------
        X : object
            Always ignored, exists for compatibility.

        y : object
            Always ignored, exists for compatibility.

        groups : object
            Always ignored, exists for compatibility.

        Returns
        -------
        n_splits : int
            Returns the number of splitting iterations in the cross-validator.
        """
        return self.n_splits


class KFold(_BaseKFold):
    """K-Folds cross-validator

    Provides train/test indices to split data in train/test sets. Split
    dataset into k consecutive folds (without shuffling by default).

    Each fold is then used once as a validation while the k - 1 remaining
    folds form the training set.

    Parameters
    ----------
    n_splits : int, default=5
        Number of folds. Must be at least 2.

        .. versionchanged:: 0.22
            ``n_splits`` default value changed from 3 to 5.

    shuffle : bool, default=False
        Whether to shuffle the data before splitting into batches.
        Note that the samples within each split will not be shuffled.

    random_state : int, RandomState instance or None, default=None
        When `shuffle` is True, `random_state` affects the ordering of the
        indices, which controls the randomness of each fold. Otherwise, this
        parameter has no effect.
        Pass an int for reproducible output across multiple function calls.
        See :term:`Glossary <random_state>`.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.learn.model_selection import KFold
    >>> X = mt.array([[1, 2], [3, 4], [1, 2], [3, 4]])
    >>> y = mt.array([1, 2, 3, 4])
    >>> kf = KFold(n_splits=2)
    >>> kf.get_n_splits(X)
    2
    >>> print(kf)
    KFold(n_splits=2, random_state=None, shuffle=False)
    >>> for train_index, test_index in kf.split(X):
    ...     print("TRAIN:", train_index, "TEST:", test_index)
    ...     X_train, X_test = X[train_index], X[test_index]
    ...     y_train, y_test = y[train_index], y[test_index]
    TRAIN: [2 3] TEST: [0 1]
    TRAIN: [0 1] TEST: [2 3]

    Notes
    -----
    The first ``n_samples % n_splits`` folds have size
    ``n_samples // n_splits + 1``, other folds have size
    ``n_samples // n_splits``, where ``n_samples`` is the number of samples.

    Randomized CV splitters may return different results for each call of
    split. You can make the results identical by setting `random_state`
    to an integer.

    See Also
    --------
    StratifiedKFold : Takes group information into account to avoid building
        folds with imbalanced class distributions (for binary or multiclass
        classification tasks).

    GroupKFold : K-fold iterator variant with non-overlapping groups.

    RepeatedKFold : Repeats K-Fold n times.
    """

    def __init__(self, n_splits=5, *, shuffle=False, random_state=None):
        super().__init__(n_splits=n_splits, shuffle=shuffle, random_state=random_state)

    def split(self, X, y=None, groups=None):
        """Generate indices to split data into training and test set.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        y : array-like of shape (n_samples,), default=None
            The target variable for supervised learning problems.

        groups : array-like of shape (n_samples,), default=None
            Group labels for the samples used while splitting the dataset into
            train/test set.

        Yields
        ------
        train : ndarray
            The training set indices for that split.

        test : ndarray
            The testing set indices for that split.
        """
        X, y, groups = indexable(X, y, groups)
        n_samples = _num_samples(X)
        if self.n_splits > n_samples:
            raise ValueError(
                (
                    "Cannot have number of splits n_splits={0} greater"
                    " than the number of samples: n_samples={1}."
                ).format(self.n_splits, n_samples)
            )

        indices = mt.arange(n_samples)
        if self.shuffle:
            check_random_state(self.random_state).shuffle(indices)

        n_splits = self.n_splits
        fold_sizes = np.full(n_splits, n_samples // n_splits, dtype=int)
        fold_sizes[: n_samples % n_splits] += 1
        current = 0
        for fold_size in fold_sizes:
            start, stop = current, current + fold_size
            train_index = mt.concatenate([indices[:start], indices[stop:]])
            yield train_index, indices[start:stop]
            current = stop



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import os
from abc import ABC, abstractmethod
from concurrent.futures import Executor, ThreadPoolExecutor
from typing import Any, Optional, Union

from ..lib.aio import AioFileObject


class StorageFileObject(AioFileObject):
    def __init__(
        self,
        file: Any,
        object_id: Any,
        loop: asyncio.BaseEventLoop = None,
        executor: Executor = None,
    ):
        self._object_id = object_id
        if executor is None:
            executor = ThreadPoolExecutor()
        super().__init__(file, loop=loop, executor=executor)

    @property
    def object_id(self):
        return self._object_id

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await super().__aexit__(exc_type, exc_val, exc_tb)
        if self._executor:
            self._executor.shutdown(wait=False)


class BufferWrappedFileObject(ABC):
    def __init__(self, object_id: Any, mode: str, size: Optional[int] = None):
        # check arguments
        assert mode in ("w", "r"), 'mode must be "w" or "r"'
        if mode == "w" and size is None:  # pragma: no cover
            raise ValueError("size must be provided to write")

        self._object_id = object_id
        self._size = size
        self._mode = mode

        self._offset = 0
        self._initialized = False
        self._closed = False

        self._mv = None
        self._buffer = None

    @abstractmethod
    def _read_init(self):
        """
        Initialization for read purpose.
        """

    @abstractmethod
    def _write_init(self):
        """
        Initialization for write purpose.
        """

    @property
    def object_id(self):
        return self._object_id

    @property
    def buffer(self):
        return self._buffer

    @property
    def mode(self):
        return self._mode

    def read(self, size=-1):
        if not self._initialized:
            self._read_init()
            self._initialized = True

        offset = self._offset
        size = self._size if size < 0 else size
        end = min(self._size, offset + size)
        result = self._mv[offset:end]
        self._offset = end
        return result

    def write(self, content: Union[bytes, memoryview]):
        if not self._initialized:
            self._write_init()
            self._initialized = True

        offset = self._offset
        content_length = getattr(content, "nbytes", len(content))
        new_offset = offset + content_length
        self._mv[offset:new_offset] = content
        self._offset = new_offset

    def seek(self, offset: int, whence: int = os.SEEK_SET):
        if not self._initialized:
            self._read_init()
            self._initialized = True

        if whence == os.SEEK_END:
            new_offset = self._size + offset
        elif whence == os.SEEK_CUR:
            new_offset = self._offset + offset
        else:
            assert whence == os.SEEK_SET
            new_offset = offset
        if new_offset < 0 or new_offset >= self._size:
            raise ValueError(f"File offset should be limited to (0, {self._size})")
        self._offset = new_offset
        return self._offset

    def tell(self):
        return self._offset

    @abstractmethod
    def _read_close(self):
        """
        Close for read.
        """

    @abstractmethod
    def _write_close(self):
        """
        Close for write.
        """

    def close(self):
        if self._closed:
            return

        self._closed = True
        if self._mode == "w":
            self._write_close()
        else:
            self._read_close()
        self._mv = None
        self._buffer = None



import sys
from functools import wraps

from . import Traceback


class Error(object):
    def __init__(self, exc_type, exc_value, traceback):
        self.exc_type = exc_type
        self.exc_value = exc_value
        self.__traceback = Traceback(traceback)

    @property
    def traceback(self):
        return self.__traceback.as_traceback()

    def reraise(self):
        raise self.exc_value.with_traceback(self.traceback) from None


def return_error(func, exc_type=Exception):
    @wraps(func)
    def return_exceptions_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except exc_type:
            return Error(*sys.exc_info())

    return return_exceptions_wrapper


returns_error = return_errors = returns_errors = return_error  # cause I make too many typos


@return_error
def apply_with_return_error(args):
    """
    args is a tuple where the first argument is a callable.
    eg::
        apply_with_return_error((func, 1, 2, 3)) - this will call func(1, 2, 3)
    """
    return args[0](*args[1:])



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Dict, List


class AbstractLifecycleAPI(ABC):
    @abstractmethod
    async def decref_tileables(
        self, tileable_keys: List[str], counts: List[int] = None
    ):
        """
        Decref tileables.

        Parameters
        ----------
        tileable_keys : list
            List of tileable keys.
        counts: list
            List of ref count.
        """

    @abstractmethod
    async def get_all_chunk_ref_counts(self) -> Dict[str, int]:
        """
        Get all chunk keys' ref counts.

        Returns
        -------
        key_to_ref_counts: dict
        """



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...core import recursive_tile
from ...lib.sparse.array import get_sparse_module, SparseNDArray
from ...serialization.serializables import KeyField, Int64Field
from ...tensor.array_utils import as_same_device, device
from ...tensor.core import TensorOrder
from ...tensor.utils import decide_unify_split
from ...utils import has_unknown_shape
from ..operands import LearnOperand, LearnOperandMixin, OutputType


class KNeighborsGraph(LearnOperand, LearnOperandMixin):
    _op_type_ = OperandDef.KNEIGHBORS_GRAPH

    _a_data = KeyField("a_data")
    _a_ind = KeyField("a_ind")
    _n_neighbors = Int64Field("n_neighbors")

    def __init__(self, a_data=None, a_ind=None, n_neighbors=None, **kw):
        super().__init__(_a_data=a_data, _a_ind=a_ind, _n_neighbors=n_neighbors, **kw)
        self.output_types = [OutputType.tensor]

    @property
    def a_data(self):
        return self._a_data

    @property
    def a_ind(self):
        return self._a_ind

    @property
    def n_neighbors(self):
        return self._n_neighbors

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self._a_data is not None:
            self._a_data = self._inputs[0]
        self._a_ind = self._inputs[-1]

    def __call__(self, A_data, A_ind, shape):
        inputs = []
        if A_data is not None:
            inputs.append(A_data)
        inputs.append(A_ind)
        return self.new_tileable(
            inputs, dtype=np.dtype(np.float64), shape=shape, order=TensorOrder.C_ORDER
        )

    @classmethod
    def tile(cls, op):
        if has_unknown_shape(*op.inputs):
            yield
        A_data, A_ind = op.a_data, op.a_ind
        out = op.outputs[0]

        shape1 = A_ind.shape[1]
        if A_data is not None:
            # mode == 'distance'
            axis0_chunk_sizes = decide_unify_split(A_data.nsplits[0], A_ind.nsplits[0])
            A_data = yield from recursive_tile(
                A_data.rechunk({0: axis0_chunk_sizes, 1: shape1})
            )
            A_ind = yield from recursive_tile(
                A_ind.rechunk({0: axis0_chunk_sizes, 1: shape1})
            )
        else:
            # mode == 'connectivity'
            A_ind = yield from recursive_tile(A_ind.rechunk({1: shape1}))

        out_chunks = []
        for i, ind_c in enumerate(A_ind.chunks):
            chunk_op = op.copy().reset_key()
            chunk_inputs = [ind_c]
            if A_data is not None:
                data_c = A_data.cix[i, 0]
                chunk_inputs.insert(0, data_c)
            out_chunk = chunk_op.new_chunk(
                chunk_inputs,
                dtype=out.dtype,
                shape=(ind_c.shape[0], out.shape[1]),
                order=out.order,
                index=(i, 0),
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = out.params
        params["chunks"] = out_chunks
        params["nsplits"] = (A_ind.nsplits[0], (out.shape[1],))
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )
        out = op.outputs[0]
        n_samples1, n_samples2 = out.shape
        n_neighbors = op.n_neighbors
        n_nonzero = n_samples1 * n_neighbors

        with device(device_id):
            A_ind = inputs[-1]
            A_indptr = xp.arange(0, n_nonzero + 1, n_neighbors)

            if op.a_data is None:
                # mode == 'connectivity
                A_data = xp.ones(n_samples1 * n_neighbors)
            else:
                # mode == 'distance'
                A_data = xp.ravel(inputs[0])

            xps = get_sparse_module(A_ind)
            graph = xps.csr_matrix(
                (A_data, A_ind.ravel(), A_indptr), shape=(n_samples1, n_samples2)
            )
            ctx[out.key] = SparseNDArray(graph)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import AnyField
from ..utils import gen_random_seeds
from .core import TensorRandomOperandMixin, handle_array, TensorDistribution


class TensorLognormal(TensorDistribution, TensorRandomOperandMixin):
    _input_fields_ = ["mean", "sigma"]
    _op_type_ = OperandDef.RAND_LOGNORMAL

    _fields_ = "mean", "sigma", "size"
    mean = AnyField("mean")
    sigma = AnyField("sigma")
    _func_name = "lognormal"

    def __call__(self, mean, sigma, chunk_size=None):
        return self.new_tensor([mean, sigma], None, raw_chunk_size=chunk_size)


def lognormal(
    random_state, mean=0.0, sigma=1.0, size=None, chunk_size=None, gpu=None, dtype=None
):
    r"""
    Draw samples from a log-normal distribution.

    Draw samples from a log-normal distribution with specified mean,
    standard deviation, and array shape.  Note that the mean and standard
    deviation are not the values for the distribution itself, but of the
    underlying normal distribution it is derived from.

    Parameters
    ----------
    mean : float or array_like of floats, optional
        Mean value of the underlying normal distribution. Default is 0.
    sigma : float or array_like of floats, optional
        Standard deviation of the underlying normal distribution. Should
        be greater than zero. Default is 1.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``mean`` and ``sigma`` are both scalars.
        Otherwise, ``np.broadcast(mean, sigma).size`` samples are drawn.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    dtype : data-type, optional
      Data-type of the returned tensor.

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized log-normal distribution.

    See Also
    --------
    scipy.stats.lognorm : probability density function, distribution,
        cumulative density function, etc.

    Notes
    -----
    A variable `x` has a log-normal distribution if `log(x)` is normally
    distributed.  The probability density function for the log-normal
    distribution is:

    .. math:: p(x) = \frac{1}{\sigma x \sqrt{2\pi}}
                     e^{(-\frac{(ln(x)-\mu)^2}{2\sigma^2})}

    where :math:`\mu` is the mean and :math:`\sigma` is the standard
    deviation of the normally distributed logarithm of the variable.
    A log-normal distribution results if a random variable is the *product*
    of a large number of independent, identically-distributed variables in
    the same way that a normal distribution results if the variable is the
    *sum* of a large number of independent, identically-distributed
    variables.

    References
    ----------
    .. [1] Limpert, E., Stahel, W. A., and Abbt, M., "Log-normal
           Distributions across the Sciences: Keys and Clues,"
           BioScience, Vol. 51, No. 5, May, 2001.
           http://stat.ethz.ch/~stahel/lognormal/bioscience.pdf
    .. [2] Reiss, R.D. and Thomas, M., "Statistical Analysis of Extreme
           Values," Basel: Birkhauser Verlag, 2001, pp. 31-32.

    Examples
    --------
    Draw samples from the distribution:

    >>> import mars.tensor as mt

    >>> mu, sigma = 3., 1. # mean and standard deviation
    >>> s = mt.random.lognormal(mu, sigma, 1000)

    Display the histogram of the samples, along with
    the probability density function:

    >>> import matplotlib.pyplot as plt
    >>> count, bins, ignored = plt.hist(s.execute(), 100, normed=True, align='mid')

    >>> x = mt.linspace(min(bins), max(bins), 10000)
    >>> pdf = (mt.exp(-(mt.log(x) - mu)**2 / (2 * sigma**2))
    ...        / (x * sigma * mt.sqrt(2 * mt.pi)))

    >>> plt.plot(x.execute(), pdf.execute(), linewidth=2, color='r')
    >>> plt.axis('tight')
    >>> plt.show()

    Demonstrate that taking the products of random samples from a uniform
    distribution can be fit well by a log-normal probability density
    function.

    >>> # Generate a thousand samples: each is the product of 100 random
    >>> # values, drawn from a normal distribution.
    >>> b = []
    >>> for i in range(1000):
    ...    a = 10. + mt.random.random(100)
    ...    b.append(mt.product(a).execute())

    >>> b = mt.array(b) / mt.min(b) # scale values to be positive
    >>> count, bins, ignored = plt.hist(b.execute(), 100, normed=True, align='mid')
    >>> sigma = mt.std(mt.log(b))
    >>> mu = mt.mean(mt.log(b))

    >>> x = mt.linspace(min(bins), max(bins), 10000)
    >>> pdf = (mt.exp(-(mt.log(x) - mu)**2 / (2 * sigma**2))
    ...        / (x * sigma * mt.sqrt(2 * mt.pi)))

    >>> plt.plot(x.execute(), pdf.execute(), color='r', linewidth=2)
    >>> plt.show()
    """
    if dtype is None:
        dtype = (
            np.random.RandomState()
            .lognormal(handle_array(mean), handle_array(sigma), size=(0,))
            .dtype
        )
    size = random_state._handle_size(size)
    seed = gen_random_seeds(1, random_state.to_numpy())[0]
    op = TensorLognormal(seed=seed, size=size, gpu=gpu, dtype=dtype)
    return op(mean, sigma, chunk_size=chunk_size)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List

from .... import oscar as mo
from ..core import NodeRole, NodeStatus
from ..locator import SupervisorLocatorActor

logger = logging.getLogger(__name__)


class WorkerSupervisorLocatorActor(SupervisorLocatorActor):
    _node_role = NodeRole.WORKER

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._node_info_ref = None

    @classmethod
    def default_uid(cls):
        return SupervisorLocatorActor.__name__

    async def _set_supervisors(self, supervisors: List[str]):
        await super()._set_supervisors(supervisors)
        if supervisors and self._node_info_ref is None:
            from ..supervisor.node_info import NodeInfoCollectorActor

            supervisor_addr = self.get_supervisor(NodeInfoCollectorActor.default_uid())
            try:
                self._node_info_ref = await mo.actor_ref(
                    uid=NodeInfoCollectorActor.default_uid(), address=supervisor_addr
                )
            except (OSError, mo.ServerClosed, mo.ActorNotExist):
                self._node_info_ref = None

    async def _get_supervisors_from_backend(self, filter_ready: bool = True):
        try:
            assert self._node_info_ref is not None
            statuses = (
                {NodeStatus.READY}
                if filter_ready
                else {NodeStatus.READY, NodeStatus.STARTING}
            )
            infos = await self._node_info_ref.get_nodes_info(
                role=NodeRole.SUPERVISOR, statuses=statuses
            )
            return list(infos)
        except (AssertionError, OSError, mo.ServerClosed, mo.ActorNotExist):
            self._node_info_ref = None
            return await self._backend.get_supervisors(filter_ready=filter_ready)

    async def _watch_supervisor_from_node_info(self):
        assert self._node_info_ref is not None
        version = None
        while True:
            version, infos = await self._node_info_ref.watch_nodes(
                role=NodeRole.SUPERVISOR, version=version
            )
            yield list(infos)

    async def _watch_supervisors_from_backend(self):
        while True:
            try:
                async for supervisors in self._watch_supervisor_from_node_info():
                    yield supervisors
            except (AssertionError, OSError, mo.ServerClosed, mo.ActorNotExist):
                self._node_info_ref = None

            async for supervisors in self._backend.watch_supervisors():
                yield supervisors
                if self._node_info_ref is not None:
                    break



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Dict, List, Any

from .... import oscar as mo
from ....core import ChunkType
from ....core.operand import Fuse
from ....lib.aio import alru_cache
from ....typing import BandType
from ....utils import get_chunk_params
from ..core import get_meta_type
from ..store import AbstractMetaStore
from ..supervisor.core import MetaStoreManagerActor, MetaStoreActor
from ..worker.core import WorkerMetaStoreManagerActor
from .core import AbstractMetaAPI


class BaseMetaAPI(AbstractMetaAPI):
    def __init__(self, session_id: str, meta_store: mo.ActorRefType[AbstractMetaStore]):
        # make sure all meta types registered
        from .. import metas

        del metas

        self._session_id = session_id
        self._meta_store = meta_store

    @mo.extensible
    async def set_tileable_meta(
        self, tileable, memory_size: int = None, store_size: int = None, **extra
    ):
        from ....dataframe.core import (
            DATAFRAME_TYPE,
            DATAFRAME_GROUPBY_TYPE,
            SERIES_GROUPBY_TYPE,
        )

        params = tileable.params.copy()
        if isinstance(
            tileable, (DATAFRAME_TYPE, DATAFRAME_GROUPBY_TYPE, SERIES_GROUPBY_TYPE)
        ):
            # dataframe needs some special process for now
            del params["columns_value"]
            del params["dtypes"]
            params.pop("key_dtypes", None)
            params["dtypes_value"] = tileable.dtypes_value
        params["nsplits"] = tileable.nsplits
        params.update(extra)
        meta = get_meta_type(type(tileable))(
            object_id=tileable.key,
            **params,
            memory_size=memory_size,
            store_size=store_size
        )
        return await self._meta_store.set_meta(tileable.key, meta)

    @mo.extensible
    async def get_tileable_meta(
        self, object_id: str, fields: List[str] = None
    ) -> Dict[str, Any]:
        return await self._meta_store.get_meta(object_id, fields=fields)

    @mo.extensible
    async def del_tileable_meta(self, object_id: str):
        return await self._meta_store.del_meta(object_id)

    @classmethod
    def _extract_chunk_meta(
        cls,
        chunk: ChunkType,
        memory_size: int = None,
        store_size: int = None,
        bands: List[BandType] = None,
        fields: List[str] = None,
        exclude_fields: List[str] = None,
        **extra
    ):
        if isinstance(chunk.op, Fuse):
            # fuse op
            chunk = chunk.chunk
        params = get_chunk_params(chunk)
        chunk_key = extra.pop("chunk_key", chunk.key)
        object_ref = extra.pop("object_ref", None)
        params.update(extra)

        if object_ref:
            object_refs = (
                [object_ref] if not isinstance(object_ref, list) else object_ref
            )
        else:
            object_refs = []

        if fields is not None:
            fields = set(fields)
            params = {k: v for k, v in params.items() if k in fields}
        elif exclude_fields is not None:
            exclude_fields = set(exclude_fields)
            params = {k: v for k, v in params.items() if k not in exclude_fields}

        return get_meta_type(type(chunk))(
            object_id=chunk_key,
            **params,
            bands=bands,
            memory_size=memory_size,
            store_size=store_size,
            object_refs=object_refs
        )

    @mo.extensible
    async def set_chunk_meta(
        self,
        chunk: ChunkType,
        memory_size: int = None,
        store_size: int = None,
        bands: List[BandType] = None,
        fields: List[str] = None,
        exclude_fields: List[str] = None,
        **extra
    ):
        """
        Parameters
        ----------
        chunk: ChunkType
            chunk to set meta
        memory_size: int
            memory size for chunk data
        store_size: int
            serialized size for chunk data
        bands:
            chunk data bands
        fields: list
            fields to include in meta
        exclude_fields: list
            fields to exclude in meta
        extra

        Returns
        -------

        """
        meta = self._extract_chunk_meta(
            chunk,
            memory_size=memory_size,
            store_size=store_size,
            bands=bands,
            fields=fields,
            exclude_fields=exclude_fields,
            **extra
        )
        return await self._meta_store.set_meta(meta.object_id, meta)

    @set_chunk_meta.batch
    async def batch_set_chunk_meta(self, args_list, kwargs_list):
        set_chunk_metas = []
        for args, kwargs in zip(args_list, kwargs_list):
            meta = self._extract_chunk_meta(*args, **kwargs)
            set_chunk_metas.append(
                self._meta_store.set_meta.delay(meta.object_id, meta)
            )
        return await self._meta_store.set_meta.batch(*set_chunk_metas)

    @mo.extensible
    async def get_chunk_meta(
        self, object_id: str, fields: List[str] = None, error="raise"
    ):
        return await self._meta_store.get_meta(object_id, fields=fields, error=error)

    @get_chunk_meta.batch
    async def batch_get_chunk_meta(self, args_list, kwargs_list):
        get_chunk_metas = []
        for args, kwargs in zip(args_list, kwargs_list):
            get_chunk_metas.append(self._meta_store.get_meta.delay(*args, **kwargs))
        return await self._meta_store.get_meta.batch(*get_chunk_metas)

    @mo.extensible
    async def del_chunk_meta(self, object_id: str):
        """
        Parameters
        ----------
        object_id: str
            chunk id
        """
        return await self._meta_store.del_meta(object_id)

    @del_chunk_meta.batch
    async def batch_del_chunk_meta(self, args_list, kwargs_list):
        del_chunk_metas = []
        for args, kwargs in zip(args_list, kwargs_list):
            del_chunk_metas.append(self._meta_store.del_meta.delay(*args, **kwargs))
        return await self._meta_store.del_meta.batch(*del_chunk_metas)

    @mo.extensible
    async def add_chunk_bands(self, object_id: str, bands: List[BandType]):
        return await self._meta_store.add_chunk_bands(object_id, bands)

    @add_chunk_bands.batch
    async def batch_add_chunk_bands(self, args_list, kwargs_list):
        add_chunk_bands_tasks = []
        for args, kwargs in zip(args_list, kwargs_list):
            add_chunk_bands_tasks.append(
                self._meta_store.add_chunk_bands.delay(*args, **kwargs)
            )
        return await self._meta_store.add_chunk_bands.batch(*add_chunk_bands_tasks)

    @mo.extensible
    async def remove_chunk_bands(self, object_id: str, bands: List[BandType]):
        return await self._meta_store.remove_chunk_bands(object_id, bands)

    @remove_chunk_bands.batch
    async def batch_remove_chunk_bands(self, args_list, kwargs_list):
        remove_chunk_bands_tasks = []
        for args, kwargs in zip(args_list, kwargs_list):
            remove_chunk_bands_tasks.append(
                self._meta_store.remove_chunk_bands.delay(*args, **kwargs)
            )
        return await self._meta_store.remove_chunk_bands.batch(
            *remove_chunk_bands_tasks
        )

    @mo.extensible
    async def get_band_chunks(self, band: BandType) -> List[str]:
        return await self._meta_store.get_band_chunks(band)


class MetaAPI(BaseMetaAPI):
    @classmethod
    @alru_cache(cache_exceptions=False)
    async def create(cls, session_id: str, address: str) -> "MetaAPI":
        """
        Create Meta API.

        Parameters
        ----------
        session_id : str
            Session ID.
        address : str
            Supervisor address.

        Returns
        -------
        meta_api
            Meta api.
        """
        meta_store_ref = await mo.actor_ref(address, MetaStoreActor.gen_uid(session_id))

        return MetaAPI(session_id, meta_store_ref)


class MockMetaAPI(MetaAPI):
    @classmethod
    async def create(cls, session_id: str, address: str) -> "MetaAPI":
        # create an Actor for mock
        try:
            meta_store_manager_ref = await mo.create_actor(
                MetaStoreManagerActor,
                "dict",
                dict(),
                address=address,
                uid=MetaStoreManagerActor.default_uid(),
            )
        except mo.ActorAlreadyExist:
            # ignore if actor exists
            meta_store_manager_ref = await mo.actor_ref(
                MetaStoreManagerActor,
                address=address,
                uid=MetaStoreManagerActor.default_uid(),
            )
        try:
            await meta_store_manager_ref.new_session_meta_store(session_id)
        except mo.ActorAlreadyExist:
            pass
        return await super().create(session_id=session_id, address=address)


class WorkerMetaAPI(BaseMetaAPI):
    @classmethod
    @alru_cache(cache_exceptions=False)
    async def create(cls, session_id: str, address: str) -> "WorkerMetaAPI":
        """
        Create worker meta API.

        Parameters
        ----------
        session_id : str
            Session ID.
        address : str
            Worker address.

        Returns
        -------
        meta_api
            Worker meta api.
        """
        worker_meta_store_manager_ref = await mo.actor_ref(
            uid=WorkerMetaStoreManagerActor.default_uid(), address=address
        )
        worker_meta_store_ref = (
            await worker_meta_store_manager_ref.new_session_meta_store(session_id)
        )
        return WorkerMetaAPI(session_id, worker_meta_store_ref)


class MockWorkerMetaAPI(WorkerMetaAPI):
    @classmethod
    async def create(cls, session_id: str, address: str) -> "WorkerMetaAPI":
        # create an Actor for mock
        try:
            await mo.create_actor(
                WorkerMetaStoreManagerActor,
                "dict",
                dict(),
                address=address,
                uid=WorkerMetaStoreManagerActor.default_uid(),
            )
        except mo.ActorAlreadyExist:
            # ignore if actor exists
            await mo.actor_ref(
                WorkerMetaStoreManagerActor,
                address=address,
                uid=WorkerMetaStoreManagerActor.default_uid(),
            )
        return await super().create(session_id, address)



# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import random
import time
import uuid

from ...session import new_session
from ...utils import to_str, calc_size_by_str
from ..utils import wait_services_ready
from .config import MarsApplicationConfig, MarsSupervisorConfig, MarsWorkerConfig

logger = logging.getLogger(__name__)


class YarnClusterClient:
    def __init__(self, skein_client, application_id, endpoint, is_client_managed=False):
        self._skein_client = skein_client
        self._is_client_managed = is_client_managed
        self._application_id = application_id
        self._endpoint = endpoint
        self._session = new_session(endpoint)

    @property
    def session(self):
        return self._session

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def application_id(self):
        return self._application_id

    def stop(self, status="SUCCEEDED"):
        import skein

        try:
            skein_client = skein.Client()
            app_client = skein_client.connect(self._application_id)
            app_client.shutdown(status=status)
            if self._is_client_managed:
                self._skein_client.close()
        except skein.ApplicationNotRunningError:
            pass


def _get_ready_container_count(app_client, svc):
    container_ids = set(
        c.yarn_container_id for c in app_client.get_containers([svc], ["RUNNING"])
    )
    prefixes = app_client.kv.get_prefix(svc)
    registered_ids = set(to_str(v).rsplit("@", 1)[-1] for v in prefixes.values())
    return len(container_ids.intersection(registered_ids))


def new_cluster(
    environment=None,
    supervisor_num=1,
    supervisor_cpu=None,
    supervisor_mem=None,
    worker_num=1,
    worker_cpu=None,
    worker_mem=None,
    worker_spill_paths=None,
    worker_cache_mem=None,
    min_worker_num=None,
    timeout=None,
    log_config=None,
    skein_client=None,
    app_name=None,
    app_queue=None,
    **kwargs,
):
    import skein
    from .supervisor import YarnSupervisorCommandRunner

    def _override_envs(src, updates):
        ret = src.copy()
        ret.update(updates)
        return ret

    if worker_cpu is None or worker_mem is None:  # pragma: no cover
        raise TypeError("`worker_cpu` and `worker_mem` must be specified")

    app_name = app_name or f"mars-app-{uuid.uuid4()}"
    supervisor_mem = calc_size_by_str(supervisor_mem, None)
    worker_mem = calc_size_by_str(worker_mem, None)

    log_when_fail = kwargs.pop("log_when_fail", False)

    supervisor_extra_modules = kwargs.pop("supervisor_extra_modules", None)
    worker_extra_modules = kwargs.pop("worker_extra_modules", None)

    cmd_tmpl = kwargs.pop("cmd_tmpl", None)

    extra_envs = kwargs.pop("extra_env", dict())
    supervisor_extra_env = _override_envs(
        extra_envs, kwargs.pop("supervisor_extra_env", dict())
    )
    worker_extra_env = _override_envs(
        extra_envs, kwargs.pop("worker_extra_env", dict())
    )

    extra_args = kwargs.pop("extra_args", "")
    supervisor_extra_args = (
        extra_args + " " + kwargs.pop("supervisor_extra_args", "")
    ).strip()
    worker_extra_args = (extra_args + " " + kwargs.pop("worker_extra_args", "")).strip()

    supervisor_log_config = kwargs.pop("supervisor_log_config", log_config)
    worker_log_config = kwargs.pop("worker_log_config", log_config)

    supervisor_config = MarsSupervisorConfig(
        instances=supervisor_num,
        environment=environment,
        cpu=supervisor_cpu,
        memory=supervisor_mem,
        modules=supervisor_extra_modules,
        env=supervisor_extra_env,
        log_config=supervisor_log_config,
        extra_args=supervisor_extra_args,
        cmd_tmpl=cmd_tmpl,
    )
    worker_config = MarsWorkerConfig(
        instances=worker_num,
        environment=environment,
        cpu=worker_cpu,
        memory=worker_mem,
        spill_dirs=worker_spill_paths,
        worker_cache_mem=worker_cache_mem,
        modules=worker_extra_modules,
        env=worker_extra_env,
        log_config=worker_log_config,
        extra_args=worker_extra_args,
        cmd_tmpl=cmd_tmpl,
    )
    app_config = MarsApplicationConfig(
        app_name,
        app_queue,
        supervisor_config=supervisor_config,
        worker_config=worker_config,
    )

    skein_client = skein_client or skein.Client()
    app_id = None
    try:
        is_client_managed = skein_client is not None
        app_id = skein_client.submit(app_config.build())

        check_start_time = time.time()
        while True:
            try:
                app_client = skein_client.connect(app_id)
                break
            except skein.ApplicationNotRunningError:  # pragma: no cover
                time.sleep(0.5)
                if timeout and time.time() - check_start_time > timeout:
                    raise

        logger.debug(
            "Application client for %s at %s retrieved", app_id, app_client.address
        )

        # wait until supervisors and expected num of workers are ready
        min_worker_num = int(min_worker_num or worker_num)
        limits = [supervisor_num, min_worker_num]
        services = [MarsSupervisorConfig.service_name, MarsWorkerConfig.service_name]

        wait_services_ready(
            services,
            limits,
            lambda svc: _get_ready_container_count(app_client, svc),
            timeout=None if not timeout else timeout - (time.time() - check_start_time),
        )
        web_endpoint_kv = app_client.kv.get_prefix(
            YarnSupervisorCommandRunner.web_service_name
        )
        web_endpoint = random.choice(
            [to_str(v).split("@", 1)[0] for v in web_endpoint_kv.values()]
        )
        return YarnClusterClient(
            skein_client,
            app_client.id,
            web_endpoint,
            is_client_managed=is_client_managed,
        )
    except:  # noqa: E722
        skein_client = skein.Client()
        try:
            if log_when_fail:
                if app_id is not None:
                    try:
                        app_client = skein_client.connect(app_id)
                        app_client.shutdown(status="FAILED")
                    except skein.ApplicationNotRunningError:
                        pass

                    try:
                        logs = skein_client.application_logs(app_id)
                        logger.error("Error when creating cluster:\n%s", logs.dumps())
                    except ValueError:
                        logger.error(
                            "Error when creating cluster and failed to get logs"
                        )
                else:
                    logger.error("Error when creating cluster and no logs from cluster")
        finally:
            if app_id is not None:
                skein_client.kill_application(app_id)
        raise



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, ABCMeta, abstractmethod
from datetime import datetime, timedelta, tzinfo
from enum import Enum
from typing import Tuple, Type

import numpy as np
import pandas as pd

from ...utils import lazy_import

cupy = lazy_import("cupy")
cudf = lazy_import("cudf")


class PrimitiveType(Enum):
    bool = 1
    int8 = 2
    int16 = 3
    int32 = 4
    int64 = 5
    uint8 = 6
    uint16 = 7
    uint32 = 8
    uint64 = 9
    float16 = 10
    float32 = 11
    float64 = 12
    bytes = 13
    string = 14
    complex64 = 24
    complex128 = 25


_primitive_type_to_valid_types = {
    PrimitiveType.bool: (bool, np.bool_),
    PrimitiveType.int8: (int, np.int8),
    PrimitiveType.int16: (int, np.int16),
    PrimitiveType.int32: (int, np.int32),
    PrimitiveType.int64: (int, np.int64),
    PrimitiveType.uint8: (int, np.uint8),
    PrimitiveType.uint16: (int, np.uint16),
    PrimitiveType.uint32: (int, np.uint32),
    PrimitiveType.uint64: (int, np.uint64),
    PrimitiveType.float16: (float, np.float16),
    PrimitiveType.float32: (float, np.float32),
    PrimitiveType.float64: (float, np.float64),
    PrimitiveType.bytes: (bytes, np.bytes_),
    PrimitiveType.string: (str, np.unicode_),
    PrimitiveType.complex64: (complex, np.complex64),
    PrimitiveType.complex128: (complex, np.complex128),
}


class AbstractFieldType(ABC):
    __slots__ = ()

    @property
    @abstractmethod
    def type_name(self) -> str:
        """
        Type name.

        Returns
        -------
        type_name : str
        """

    @property
    def name(self) -> str:
        """
        Name of field type instance.

        Returns
        -------
        name : str
        """
        return self.type_name.capitalize()

    @property
    @abstractmethod
    def valid_types(self) -> Tuple[Type, ...]:
        """
        Valid types.

        Returns
        -------
        valid_types: tuple
            Valid types.
        """

    def validate(self, value):
        if value is not None and not isinstance(value, self.valid_types):
            raise TypeError(
                f"value needs to be instance "
                f"of {self.valid_types}, got {type(value)}"
            )

    def __call__(self, *args, **kwargs):
        return type(self)(*args, **kwargs)


class SingletonFieldType(AbstractFieldType, metaclass=ABCMeta):
    __slots__ = ()

    _instance = None

    def __new__(cls, *args, **kw):
        if cls._instance is None:
            inst = super().__new__(cls, *args, **kw)
            cls._instance = inst
        return cls._instance


class PrimitiveFieldType(AbstractFieldType):
    __slots__ = ("type",)

    _type_to_instances = dict()

    def __new__(cls, *args, **kwargs):
        primitive_type = args[0]
        try:
            return cls._type_to_instances[primitive_type]
        except KeyError:
            inst = cls._type_to_instances[primitive_type] = super().__new__(cls)
            return inst

    def __init__(self, primitive_type: PrimitiveType):
        self.type = primitive_type

    @property
    def type_name(self) -> str:
        return self.type.name

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        return _primitive_type_to_valid_types[self.type]


class SliceType(SingletonFieldType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "slice"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        return (slice,)


class NDArrayType(SingletonFieldType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "ndarray"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        if cupy is None:
            return (np.ndarray,)
        else:
            return np.ndarray, cupy.ndarray


class DtypeType(SingletonFieldType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "dtype"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        return np.dtype, pd.api.extensions.ExtensionDtype


class KeyType(SingletonFieldType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "dtype"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        from ...core.entity import ENTITY_TYPE

        return ENTITY_TYPE


class DatetimeType(SingletonFieldType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "datetime"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        return datetime, pd.Timestamp


class TimedeltaType(SingletonFieldType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "timedelta"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        return timedelta, pd.Timedelta


class IndexType(SingletonFieldType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "index"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        if cudf is None:
            return (pd.Index,)
        else:
            return pd.Index, cudf.Index


class SeriesType(SingletonFieldType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "series"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        if cudf is None:
            return (pd.Series,)
        else:
            return pd.Series, cudf.Series


class DataFrameType(SingletonFieldType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "dataframe"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        if cudf is None:
            return (pd.DataFrame,)
        else:
            return pd.DataFrame, cudf.DataFrame


class FunctionType(SingletonFieldType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "function"

    @property
    def valid_types(self) -> Tuple[Type, ...]:  # pragma: no cover
        return ()

    def validate(self, value):
        if value is not None and not callable(value):
            raise TypeError(f"value should be a function, got {type(value)}")


class NamedtupleType(SingletonFieldType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "namedtuple"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        return (tuple,)

    def validate(self, value):
        if not (isinstance(value, self.valid_types) and hasattr(value, "_fields")):
            raise TypeError(
                f"value should be instance of namedtuple, got {type(value)}"
            )


class TZInfoType(SingletonFieldType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "tzinfo"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        return (tzinfo,)


class IntervalArrayType(SingletonFieldType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "interval_array"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        return (pd.arrays.IntervalArray,)


class AnyType(SingletonFieldType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "any"

    @property
    def valid_types(self) -> Tuple[Type, ...]:  # pragma: no cover
        return ()

    def validate(self, value):
        # any type is valid
        return


class _CollectionType(AbstractFieldType, metaclass=ABCMeta):
    __slots__ = ("_field_types",)

    def __init__(self, *field_types):
        self._field_types = field_types
        if len(field_types) == 0:
            self._field_types = (AnyType(), Ellipsis)

    @property
    def name(self) -> str:
        base_name = super().name
        if self.is_homogeneous():
            if isinstance(self._field_types[0], AnyType):
                return base_name
            else:
                return f"{base_name}[{self._field_types[0].name}, ...]"
        else:
            field_type_names = ", ".join([ft.name for ft in self._field_types])
            return f"{base_name}[{field_type_names}]"

    def is_homogeneous(self):
        return len(self._field_types) == 1 or (
            len(self._field_types) == 2 and self._field_types[1] is Ellipsis
        )

    def validate(self, value):
        if value is None:
            return
        if not isinstance(value, self.valid_types):
            raise TypeError(
                f"value should be instance of {self.valid_types}, got {type(value)}"
            )
        if self.is_homogeneous():
            field_type: AbstractFieldType = self._field_types[0]
            if not isinstance(field_type, AnyType):
                for item in value:
                    try:
                        field_type.validate(item)
                    except TypeError:
                        raise TypeError(
                            f"item should be instance of "
                            f"{field_type.valid_types}, "
                            f"got {type(item)}"
                        )
        else:
            if len(value) != len(self._field_types):
                raise ValueError(
                    f"value should own {len(self._field_types)} items, "
                    f"got {len(value)} items"
                )
            for expect_field_type, item in zip(self._field_types, value):
                try:
                    expect_field_type.validate(item)
                except TypeError:
                    raise TypeError(
                        f"item should be instance of "
                        f"{expect_field_type.valid_types}, "
                        f"got {type(item)}"
                    )


class ListType(_CollectionType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "list"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        return (list,)


class TupleType(_CollectionType):
    __slots__ = ()

    @property
    def type_name(self) -> str:
        return "tuple"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        return (tuple,)


class DictType(AbstractFieldType):
    __slots__ = "key_type", "value_type"

    key_type: AbstractFieldType
    value_type: AbstractFieldType

    def __init__(
        self, key_type: AbstractFieldType = None, value_type: AbstractFieldType = None
    ):
        if key_type is None:
            key_type = AnyType()
        if value_type is None:
            value_type = AnyType()
        self.key_type = key_type
        self.value_type = value_type

    @property
    def type_name(self) -> str:
        return "dict"

    @property
    def name(self) -> str:
        if isinstance(self.key_type, AnyType) and isinstance(self.value_type, AnyType):
            return "Dict"
        else:
            return f"Dict[{self.key_type.name}, {self.value_type.name}]"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        return (dict,)

    def validate(self, value):
        super().validate(value)
        if value is None:
            return
        for k, v in value.items():
            try:
                self.key_type.validate(k)
            except TypeError:
                raise TypeError(
                    f"key should be instance of "
                    f"{self.key_type.valid_types}, got {type(k)}"
                )
            try:
                self.value_type.validate(v)
            except TypeError:
                raise TypeError(
                    f"value should be instance of "
                    f"{self.value_type.valid_types}, got {type(v)}"
                )


class ReferenceType(AbstractFieldType):
    __slots__ = ("reference_type",)

    reference_type: Type

    def __init__(self, reference_type: Type = None):
        if reference_type is None:
            reference_type = object
        self.reference_type = reference_type

    @property
    def type_name(self) -> str:
        return "reference"

    @property
    def valid_types(self) -> Tuple[Type, ...]:
        return (self.reference_type,)


class FieldTypes:
    # primitive type
    bool = PrimitiveFieldType(PrimitiveType.bool)
    int8 = PrimitiveFieldType(PrimitiveType.int8)
    int16 = PrimitiveFieldType(PrimitiveType.int16)
    int32 = PrimitiveFieldType(PrimitiveType.int32)
    int64 = PrimitiveFieldType(PrimitiveType.int64)
    uint8 = PrimitiveFieldType(PrimitiveType.uint8)
    uint16 = PrimitiveFieldType(PrimitiveType.uint16)
    uint32 = PrimitiveFieldType(PrimitiveType.uint32)
    uint64 = PrimitiveFieldType(PrimitiveType.uint64)
    float16 = PrimitiveFieldType(PrimitiveType.float16)
    float32 = PrimitiveFieldType(PrimitiveType.float32)
    float64 = PrimitiveFieldType(PrimitiveType.float64)
    complex64 = PrimitiveFieldType(PrimitiveType.complex64)
    complex128 = PrimitiveFieldType(PrimitiveType.complex128)
    bytes = PrimitiveFieldType(PrimitiveType.bytes)
    string = PrimitiveFieldType(PrimitiveType.string)

    key = KeyType()

    # Python types
    slice = SliceType()
    datetime = DatetimeType()
    # alias of datetime
    datatime64 = DatetimeType()
    timedelta = TimedeltaType()
    # alias of timedelta
    timedelta64 = TimedeltaType()
    tzinfo = TZInfoType()
    function = FunctionType()
    namedtuple = NamedtupleType()
    reference = ReferenceType()
    any = AnyType()
    # equivalent to any
    pickled = AnyType()

    # collection
    list = ListType()
    tuple = TupleType()
    dict = DictType()

    # numpy
    ndarray = NDArrayType()
    # alias of ndarray
    arr = NDArrayType()
    dtype = DtypeType()

    # pandas
    index = IndexType()
    series = SeriesType()
    dataframe = DataFrameType()
    interval_array = IntervalArrayType()
    # alias of interval_array
    interval_arr = IntervalArrayType()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes
from ...core import OutputType
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index


class DataFrameTranspose(DataFrameOperand, DataFrameOperandMixin):
    _op_code_ = opcodes.TRANSPOSE

    def __init__(self, **kw):
        super().__init__(**kw)
        self.output_types = [OutputType.dataframe]

    def __call__(self, args):
        arg = args[0]
        new_shape = arg.shape[::-1]
        columns_value = arg.index_value
        index_value = parse_index(arg.dtypes.index)
        return self.new_dataframe(
            [arg],
            shape=new_shape,
            dtypes=None,
            columns_value=columns_value,
            index_value=index_value,
        )

    @classmethod
    def tile(cls, op):
        out_chunks = []
        for c in op.inputs[0].chunks:
            chunk_op = op.copy().reset_key()
            chunk_shape = tuple(s if np.isnan(s) else int(s) for s in c.shape[::-1])
            chunk_idx = c.index[::-1]
            index_value = parse_index(c.dtypes.index)
            columns_value = c.index_value
            out_chunk = chunk_op.new_chunk(
                [c],
                shape=chunk_shape,
                index=chunk_idx,
                index_value=index_value,
                columns_value=columns_value,
                dtypes=None,
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        nsplits = op.inputs[0].nsplits[::-1]
        params = op.outputs[0].params
        return new_op.new_dataframe(
            op.inputs, chunks=out_chunks, nsplits=nsplits, **params
        )

    @classmethod
    def execute(cls, ctx, op):
        inp = ctx[op.inputs[0].key]
        ctx[op.outputs[0].key] = inp.transpose()


def transpose(*args):
    """
    Transpose index and columns.

    Reflect the DataFrame over its main diagonal by writing rows as columns
    and vice-versa. The property :attr:`.T` is an accessor to the method
    :meth:`transpose`.

    Parameters
    ----------
    *args : tuple, optional
            Accepted for compatibility with NumPy.

    Returns
    -------
    DataFrame
        The transposed DataFrame.

    See Also
    --------
    numpy.transpose : Permute the dimensions of a given array.

    Notes
    -----
    Transposing a DataFrame with mixed dtypes will result in a homogeneous
    DataFrame with the `object` dtype.

    Examples
    --------
    **Square DataFrame with homogeneous dtype**

    >>> import mars.dataframe as md
    >>> d1 = {'col1': [1, 2], 'col2': [3, 4]}
    >>> df1 = md.DataFrame(data=d1).execute()
    >>> df1
        col1  col2
    0     1     3
    1     2     4

    >>> df1_transposed = df1.T.execute() # or df1.transpose().execute()
    >>> df1_transposed
          0  1
    col1  1  2
    col2  3  4

    When the dtype is homogeneous in the original DataFrame, we get a
    transposed DataFrame with the same dtype:

    >>> df1.dtypes
    col1    int64
    col2    int64
    dtype: object

    >>> df1_transposed.dtypes
    0    int64
    1    int64
    dtype: object

    **Non-square DataFrame with mixed dtypes**

    >>> d2 = {'name': ['Alice', 'Bob'],
    ...       'score': [9.5, 8],
    ...       'employed': [False, True],
    ...       'kids': [0, 0]}
    >>> df2 = md.DataFrame(data=d2).execute()
    >>> df2
        name  score  employed  kids
    0  Alice    9.5     False     0
    1    Bob    8.0      True     0

    >>> df2_transposed = df2.T.execute() # or df2.transpose().execute()
    >>> df2_transposed
                  0     1
    name      Alice   Bob
    score       9.5   8.0
    employed  False  True
    kids          0     0

    When the DataFrame has mixed dtypes, we get a transposed DataFrame with
    the `object` dtype:

    >>> df2.dtypes
    name         object
    score       float64
    employed       bool
    kids          int64
    dtype: object

    >>> df2_transposed.dtypes
    0    object
    1    object
    dtype: object
    """
    op = DataFrameTranspose()
    return op(args)
